# BSD 2-Clause License
#
# Apprise - Push Notification Library.
# Copyright (c) 2025, Chris Caron <lead2gold@gmail.com>
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice,
#    this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

# 1. Create a BlueSky account
# 2. Access Settings -> Privacy and Security
# 3. Generate an App Password.  Optionally grant yourself access to Direct
#    Messages if you want to be able to send them
# 4. Assemble your Apprise URL like:
#       bluesky://handle@you-token-here
#
from datetime import datetime, timedelta, timezone
import json
import re

import requests

from ..attachment.base import AttachBase
from ..common import NotifyType
from ..locale import gettext_lazy as _
from ..url import PrivacyMode
from .base import NotifyBase

# For parsing handles
HANDLE_HOST_PARSE_RE = re.compile(r"(?P<handle>[^.]+)\.+(?P<host>.+)$")

IS_USER = re.compile(r"^\s*@?(?P<user>[A-Z0-9_]+)(\.+(?P<host>.+))?$", re.I)


class NotifyBlueSky(NotifyBase):
    """A wrapper for BlueSky Notifications."""

    # The default descriptive name associated with the Notification
    service_name = "BlueSky"

    # The services URL
    service_url = "https://bluesky.us/"

    # Protocol
    secure_protocol = ("bsky", "bluesky")

    # A URL that takes you to the setup/help of the specific protocol
    setup_url = "https://github.com/caronc/apprise/wiki/Notify_bluesky"

    # Support attachments
    attachment_support = True

    # XRPC Suffix URLs; Structured as:
    #  https://host/{suffix}

    # Taken right from google.auth.helpers:
    clock_skew = timedelta(seconds=10)

    # 1 hour in seconds (the lifetime of our token)
    access_token_lifetime_sec = timedelta(seconds=3600)

    # Detect your Decentralized Identitifer (DID), then you can get your Auth
    # Token.
    xrpc_suffix_did = "/xrpc/com.atproto.identity.resolveHandle"
    xrpc_suffix_session = "/xrpc/com.atproto.server.createSession"
    xrpc_suffix_record = "/xrpc/com.atproto.repo.createRecord"
    xrpc_suffix_blob = "/xrpc/com.atproto.repo.uploadBlob"
    plc_directory = "https://plc.directory/{did}"

    # BlueSky is kind enough to return how many more requests we're allowed to
    # continue to make within it's header response as:
    # RateLimit-Reset: The epoc time (in seconds) we can expect our
    #                   rate-limit to be reset.
    # RateLimit-Remaining: an integer identifying how many requests we're
    #                      still allow to make.
    request_rate_per_sec = 0

    # For Tracking Purposes
    ratelimit_reset = datetime.now(timezone.utc).replace(tzinfo=None)

    # Remaining messages
    ratelimit_remaining = 1

    # The default BlueSky host to use if one isn't specified
    bluesky_default_host = "bsky.social"

    # Our message body size
    body_maxlen = 280

    # BlueSky does not support a title
    title_maxlen = 0

    # Define object templates
    templates = ("{schema}://{user}@{password}",)

    # Define our template tokens
    template_tokens = dict(
        NotifyBase.template_tokens,
        **{
            "user": {
                "name": _("Username"),
                "type": "string",
                "required": True,
            },
            "password": {
                "name": _("Password"),
                "type": "string",
                "private": True,
                "required": True,
            },
        },
    )

    def __init__(self, **kwargs):
        """Initialize BlueSky Object."""
        super().__init__(**kwargs)

        # Our access token
        self.__access_token = self.store.get("access_token")
        self.__refresh_token = None
        self.__access_token_expiry = datetime.now(timezone.utc)
        self.__endpoint = self.store.get("endpoint")

        if not self.user:
            msg = "A BlueSky UserID/Handle must be specified."
            self.logger.warning(msg)
            raise TypeError(msg)

        # Set our default host
        self.host = self.bluesky_default_host
        self.__endpoint = (
            f"https://{self.host}" if not self.host else self.__endpoint
        )

        # Identify our Handle (if define)
        results = HANDLE_HOST_PARSE_RE.match(self.user)
        if results:
            self.user = results.group("handle").strip()
            self.host = results.group("host").strip()

        return

    def send(
        self,
        body,
        title="",
        notify_type=NotifyType.INFO,
        attach=None,
        **kwargs,
    ):
        """Perform BlueSky Notification."""

        if not self.__access_token and not self.login():
            # We failed to authenticate - we're done
            return False

        # Track our returning blob IDs as they're stored on the BlueSky server
        blobs = []

        if attach and self.attachment_support:
            url = f"{self.__endpoint}{self.xrpc_suffix_blob}"
            # We need to upload our payload first so that we can source it
            # in remaining messages
            for no, attachment in enumerate(attach, start=1):

                # Perform some simple error checking
                if not attachment:
                    # We could not access the attachment
                    self.logger.error(
                        "Could not access attachment"
                        f" {attachment.url(privacy=True)}."
                    )
                    return False

                if not re.match(r"^image/.*", attachment.mimetype, re.I):
                    # Only support images at this time
                    self.logger.warning(
                        "Ignoring unsupported BlueSky attachment"
                        f" {attachment.url(privacy=True)}."
                    )
                    continue

                self.logger.debug(
                    "Preparing BlueSky attachment"
                    f" {attachment.url(privacy=True)}"
                )

                # Upload our image and get our blob associated with it
                postokay, response = self._fetch(
                    url,
                    payload=attachment,
                )

                if not postokay:
                    # We can't post our attachment
                    return False

                # Prepare our filename
                filename = (
                    attachment.name if attachment.name else f"file{no:03}.dat"
                )

                if not (isinstance(response, dict) and response.get("blob")):
                    self.logger.debug(
                        "Could not attach the file to BlueSky: %s (mime=%s)",
                        filename,
                        attachment.mimetype,
                    )
                    continue

                blobs.append((response.get("blob"), filename))

        # Prepare our URL
        did, endpoint = self.get_identifier()
        url = f"{endpoint}{self.xrpc_suffix_record}"

        # prepare our batch of payloads to create
        payloads = []

        payload = {
            "collection": "app.bsky.feed.post",
            "repo": did,
            "record": {
                "text": body,
                # 'YYYY-mm-ddTHH:MM:SSZ'
                "createdAt": datetime.now(tz=timezone.utc).strftime("%FT%XZ"),
                "$type": "app.bsky.feed.post",
            },
        }

        if blobs:
            for no, blob in enumerate(blobs, start=1):
                _payload = payload.copy()
                if no > 1:
                    #
                    # multiple instances
                    #
                    # 1. update createdAt time
                    # 2. Change text to identify image no
                    _payload["record"]["createdAt"] = datetime.now(
                        tz=timezone.utc
                    ).strftime("%FT%XZ")
                    _payload["record"]["text"] = f"{no:02d}/{len(blobs):02d}"

                _payload["record"]["embed"] = {
                    "images": [{
                        "image": blob[0],
                        "alt": blob[1],
                    }],
                    "$type": "app.bsky.embed.images",
                }
                payloads.append(_payload)
        else:
            payloads.append(payload)

        for payload in payloads:
            # Send Login Information
            postokay, response = self._fetch(
                url,
                payload=json.dumps(payload),
            )
            if not postokay:
                # We failed
                # Bad responses look like:
                # {
                #   'error': 'InvalidRequest',
                #   'message': 'reason'
                # }
                return False
        return True

    def get_identifier(self, user=None, login=False):
        """Performs a Decentralized User Lookup and returns the identifier."""

        if user is None:
            user = self.user

        user = f"{user}.{self.host}" if "." not in user else f"{user}"
        did_key = f"did.{user}"
        endpoint_key = f"endpoint.{user}"

        did = self.store.get(did_key)
        endpoint = self.store.get(endpoint_key)
        if did and endpoint:
            # Early return
            return did, endpoint

        # Step 1: Acquire DID from bsky.app
        url = f"https://public.api.bsky.app{self.xrpc_suffix_did}"
        params = {"handle": user}

        # Send Login Information
        postokay, response = self._fetch(
            url,
            params=params,
            method="GET",
            # We set this boolean so internal recursion doesn't take place.
            login=login,
        )

        if not postokay or not response or "did" not in response:
            # We failed
            return (False, False)

        # Store our DID
        did = response.get("did")

        # Step 2: Use DID to find the PDS
        if did.startswith("did:plc:"):
            pds_url = self.plc_directory.format(did=did)

            # PDS Query
            postokay, service_response = self._fetch(
                pds_url,
                method="GET",
                # We set this boolean so internal recursion doesn't take place.
                login=login,
            )
            if (
                not postokay
                or not service_response
                or "service" not in service_response
            ):
                # We failed
                return (False, False)

            endpoint = next(
                (
                    s["serviceEndpoint"]
                    for s in service_response.get("service", [])
                    if s["type"] == "AtprotoPersonalDataServer"
                ),
                None,
            )

        elif did.startswith("did:web:"):
            # Convert to domain
            domain = did[8:]
            web_did_url = f"https://{domain}/.well-known/did.json"
            postokay, service_response = self._fetch(
                web_did_url,
                method="GET",
                # We set this boolean so internal recursion doesn't take place.
                login=login,
            )
            if (
                not postokay
                or not service_response
                or "service" not in service_response
            ):
                # We failed
                self.logger.warning(
                    "Could not fetch DID document for did:web identity "
                    f"{did}; ensure {web_did_url} is available."
                )
                return (False, False)

            endpoint = next(
                (
                    s["serviceEndpoint"]
                    for s in service_response.get("service", [])
                    if s["type"] == "AtprotoPersonalDataServer"
                ),
                None,
            )

        else:
            self.logger.warning(
                f"Unknown BlueSky DID scheme detected in {did}"
            )
            return (False, False)

        # Step 3: Send to correct endpoint
        if not endpoint:
            self.logger.warning("Failed to resolve BlueSky PDS endpoint")
            return (False, False)

        self.store.set(did_key, did)
        self.store.set(endpoint_key, endpoint)
        return (did, endpoint)

    def login(self):
        """A simple wrapper to authenticate with the BlueSky Server."""

        # Acquire our Decentralized Identitifer
        did, self.__endpoint = self.get_identifier(self.user, login=True)
        if not did:
            return False

        url = f"{self.__endpoint}{self.xrpc_suffix_session}"

        payload = {
            "identifier": did,
            "password": self.password,
        }

        # Send Login Information
        postokay, response = self._fetch(
            url,
            payload=json.dumps(payload),
            # We set this boolean so internal recursion doesn't take place.
            login=True,
        )

        # Our response object looks like this (content has been altered for
        # presentation purposes):
        # {
        #  'did': 'did:plc:ruk414jakghak402j1jqekj2',
        #  'didDoc': {
        #    '@context': [
        #      'https://www.w3.org/ns/did/v1',
        #      'https://w3id.org/security/multikey/v1',
        #      'https://w3id.org/security/suites/secp256k1-2019/v1'
        #    ],
        #    'id': 'did:plc:ruk414jakghak402j1jqekj2',
        #    'alsoKnownAs': ['at://apprise.bsky.social'],
        #    'verificationMethod': [
        #      {
        #        'id': 'did:plc:ruk414jakghak402j1jqekj2#atproto',
        #        'type': 'Multikey',
        #        'controller': 'did:plc:ruk414jakghak402j1jqekj2',
        #        'publicKeyMultibase' 'redacted'
        #      }
        #    ],
        #  'service': [
        #      {
        #        'id': '#atproto_pds',
        #        'type': 'AtprotoPersonalDataServer',
        #        'serviceEndpoint':
        #           'https://woodtuft.us-west.host.bsky.network'
        #      }
        #    ]
        #  },
        #  'handle': 'apprise.bsky.social',
        #  'email': 'whoami@gmail.com',
        #  'emailConfirmed': True,
        #  'emailAuthFactor': False,
        #  'accessJwt': 'redacted',
        #  'refreshJwt': 'redacted',
        #  'active': True,
        # }

        if not postokay or not response:
            # We failed
            return False

        # Acquire our Token
        self.__access_token = response.get("accessJwt")

        # Handle other optional arguments we can use
        self.__access_token_expiry = (
            self.access_token_lifetime_sec
            + datetime.now(timezone.utc)
            - self.clock_skew
        )

        # The Refresh Token
        self.__refresh_token = response.get("refreshJwt", self.__refresh_token)
        self.store.set(
            "access_token", self.__access_token, self.__access_token_expiry
        )
        self.store.set(
            "refresh_token", self.__refresh_token, self.__access_token_expiry
        )
        self.store.set("endpoint", self.__endpoint)

        self.logger.info(
            f"Authenticated to BlueSky as {self.user}.{self.host}"
        )
        return True

    def _fetch(
        self,
        url,
        payload=None,
        params=None,
        method="POST",
        content_type=None,
        login=False,
    ):
        """Wrapper to BlueSky API requests object."""

        # use what was specified, otherwise build headers dynamically
        headers = {
            "User-Agent": self.app_id,
            "Content-Type": (
                payload.mimetype
                if isinstance(payload, AttachBase)
                else (
                    "application/x-www-form-urlencoded; charset=utf-8"
                    if method == "GET"
                    else "application/json"
                )
            ),
        }

        if self.__access_token:
            # Set our token
            headers["Authorization"] = f"Bearer {self.__access_token}"

        # Some Debug Logging
        self.logger.debug(
            f"BlueSky {method} URL:"
            f" {url} (cert_verify={self.verify_certificate})"
        )
        self.logger.debug(
            "BlueSky Payload: %s",
            (
                str(payload)
                if not isinstance(payload, AttachBase)
                else "attach: " + payload.name
            ),
        )

        # By default set wait to None
        wait = None

        if self.ratelimit_remaining == 0:
            # Determine how long we should wait for or if we should wait at
            # all. This isn't fool-proof because we can't be sure the client
            # time (calling this script) is completely synced up with the
            # Twitter server.  One would hope we're on NTP and our clocks are
            # the same allowing this to role smoothly:

            now = datetime.now(timezone.utc).replace(tzinfo=None)
            if now < self.ratelimit_reset:
                # We need to throttle for the difference in seconds
                # We add 0.3 seconds to the end just to allow a grace
                # period.
                wait = (self.ratelimit_reset - now).total_seconds() + 0.3

        # Always call throttle before any remote server i/o is made;
        self.throttle(wait=wait)

        # Initialize a default value for our content value
        content = {}

        # acquire our request mode
        fn = requests.post if method == "POST" else requests.get
        try:
            r = fn(
                url,
                data=(
                    payload
                    if not isinstance(payload, AttachBase)
                    else payload.open()
                ),
                params=params,
                headers=headers,
                verify=self.verify_certificate,
                timeout=self.request_timeout,
            )

            # Get our JSON content if it's possible
            try:
                content = json.loads(r.content)

            except (TypeError, ValueError, AttributeError):
                # TypeError = r.content is not a String
                # ValueError = r.content is Unparsable
                # AttributeError = r.content is None
                content = {}

            # Rate limit handling... our header objects at this point are:
            # 'RateLimit-Limit': '10',     # Total # of requests per hour
            # 'RateLimit-Remaining': '9',  # Requests remaining
            # 'RateLimit-Reset': '1741631362',  # Epoch Time
            # 'RateLimit-Policy': '10;w=86400' # NoEntries;w=<window>
            try:
                # Capture rate limiting if possible
                self.ratelimit_remaining = int(
                    r.headers.get("ratelimit-remaining")
                )
                self.ratelimit_reset = datetime.fromtimestamp(
                    int(r.headers.get("ratelimit-reset")), timezone.utc
                ).replace(tzinfo=None)

            except (TypeError, ValueError):
                # This is returned if we could not retrieve this information
                # gracefully accept this state and move on
                pass

            if r.status_code != requests.codes.ok:
                # We had a problem
                status_str = NotifyBlueSky.http_response_code_lookup(
                    r.status_code
                )

                self.logger.warning(
                    "Failed to send BlueSky {} to {}: {}error={}.".format(
                        method, url, ", " if status_str else "", r.status_code
                    )
                )

                self.logger.debug(f"Response Details:\r\n{r.content}")

                # Mark our failure
                return (False, content)

        except requests.RequestException as e:
            self.logger.warning(
                f"Exception received when sending BlueSky {method} to {url}: "
            )
            self.logger.debug(f"Socket Exception: {e!s}")

            # Mark our failure
            return (False, content)

        except OSError as e:
            self.logger.warning(
                "An I/O error occurred while handling {}.".format(
                    payload.name
                    if isinstance(payload, AttachBase)
                    else payload
                )
            )
            self.logger.debug(f"I/O Exception: {e!s}")
            return (False, content)

        return (True, content)

    @property
    def url_identifier(self):
        """Returns all of the identifiers that make this URL unique from
        another simliar one.

        Targets or end points should never be identified here.
        """
        return (
            self.secure_protocol[0],
            self.user,
            self.password,
        )

    def url(self, privacy=False, *args, **kwargs):
        """Returns the URL built dynamically based on specified arguments."""

        # Apply our other parameters
        params = self.url_parameters(privacy=privacy, *args, **kwargs)

        user = self.user
        if self.host != self.bluesky_default_host:
            user += f".{self.host}"

        # our URL
        return "{schema}://{user}@{password}?{params}".format(
            schema=self.secure_protocol[0],
            user=NotifyBlueSky.quote(user, safe=""),
            password=self.pprint(
                self.password, privacy, mode=PrivacyMode.Secret, safe=""
            ),
            params=NotifyBlueSky.urlencode(params),
        )

    @staticmethod
    def parse_url(url):
        """Parses the URL and returns enough arguments that can allow us to re-
        instantiate this object."""
        results = NotifyBase.parse_url(url, verify_host=False)
        if not results:
            # We're done early as we couldn't load the results
            return results

        if not results.get("password") and results["host"]:
            results["password"] = NotifyBlueSky.unquote(results["host"])

        # Do not use host field
        results["host"] = None
        return results



# BSD 2-Clause License
#
# Apprise - Push Notification Library.
# Copyright (c) 2025, Chris Caron <lead2gold@gmail.com>
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice,
#    this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

# API reference: https://pushy.me/docs/api/send-notifications
from itertools import chain
from json import dumps, loads
import re

import requests

from ..common import NotifyType
from ..locale import gettext_lazy as _
from ..utils.parse import parse_list, validate_regex
from .base import NotifyBase

# Used to detect a Device and Topic
VALIDATE_DEVICE = re.compile(r"^@(?P<device>[a-z0-9]+)$", re.I)
VALIDATE_TOPIC = re.compile(r"^[#]?(?P<topic>[a-z0-9]+)$", re.I)

# Extend HTTP Error Messages
PUSHY_HTTP_ERROR_MAP = {
    401: "Unauthorized - Invalid Token.",
}


class NotifyPushy(NotifyBase):
    """A wrapper for Pushy Notifications."""

    # The default descriptive name associated with the Notification
    service_name = "Pushy"

    # The services URL
    service_url = "https://pushy.me/"

    # All Pushy requests are secure
    secure_protocol = "pushy"

    # A URL that takes you to the setup/help of the specific protocol
    setup_url = "https://github.com/caronc/apprise/wiki/Notify_pushy"

    # Pushy uses the http protocol with JSON requests
    notify_url = "https://api.pushy.me/push?api_key={apikey}"

    # The maximum allowable characters allowed in the body per message
    body_maxlen = 4096

    # Define object templates
    templates = ("{schema}://{apikey}/{targets}",)

    # Define our template tokens
    template_tokens = dict(
        NotifyBase.template_tokens,
        **{
            "apikey": {
                "name": _("Secret API Key"),
                "type": "string",
                "private": True,
                "required": True,
            },
            "target_device": {
                "name": _("Target Device"),
                "type": "string",
                "prefix": "@",
                "map_to": "targets",
            },
            "target_topic": {
                "name": _("Target Topic"),
                "type": "string",
                "prefix": "#",
                "map_to": "targets",
            },
            "targets": {
                "name": _("Targets"),
                "type": "list:string",
                "required": True,
            },
        },
    )

    # Define our template arguments
    template_args = dict(
        NotifyBase.template_args,
        **{
            "sound": {
                # Specify something like ping.aiff
                "name": _("Sound"),
                "type": "string",
            },
            "badge": {
                "name": _("Badge"),
                "type": "int",
                "min": 0,
            },
            "to": {
                "alias_of": "targets",
            },
            "key": {
                "alias_of": "apikey",
            },
        },
    )

    def __init__(self, apikey, targets=None, sound=None, badge=None, **kwargs):
        """Initialize Pushy Object."""
        super().__init__(**kwargs)

        # Access Token (associated with project)
        self.apikey = validate_regex(apikey)
        if not self.apikey:
            msg = f"An invalid Pushy Secret API Key ({apikey}) was specified."
            self.logger.warning(msg)
            raise TypeError(msg)

        # Get our targets
        self.devices = []
        self.topics = []

        for target in parse_list(targets):
            result = VALIDATE_TOPIC.match(target)
            if result:
                self.topics.append(result.group("topic"))
                continue

            result = VALIDATE_DEVICE.match(target)
            if result:
                self.devices.append(result.group("device"))
                continue

            self.logger.warning(
                f"Dropped invalid topic/device  ({target}) specified.",
            )

        # Setup our sound
        self.sound = sound

        # Badge
        try:
            # Acquire our badge count if we can:
            #  - We accept both the integer form as well as a string
            #    representation
            self.badge = int(badge)
            if self.badge < 0:
                raise ValueError()

        except TypeError:
            # NoneType means use Default; this is an okay exception
            self.badge = None

        except ValueError:
            self.badge = None
            self.logger.warning(
                "The specified Pushy badge ({}) is not valid ", badge
            )

        return

    def send(self, body, title="", notify_type=NotifyType.INFO, **kwargs):
        """Perform Pushy Notification."""

        if len(self.topics) + len(self.devices) == 0:
            # There were no services to notify
            self.logger.warning("There were no Pushy targets to notify.")
            return False

        # error tracking (used for function return)
        has_error = False

        # Default Header
        headers = {
            "User-Agent": self.app_id,
            "Content-Type": "application/json",
            "Accepts": "application/json",
        }

        # Our URL
        notify_url = self.notify_url.format(apikey=self.apikey)

        # Default content response object
        content = {}

        # Create a copy of targets (topics and devices)
        targets = list(self.topics) + list(self.devices)
        while len(targets):
            target = targets.pop(0)

            # prepare JSON Object
            payload = {
                # Mandatory fields
                "to": target,
                "data": {
                    "message": body,
                },
                "notification": {
                    "body": body,
                },
            }

            # Optional payload items
            if title:
                payload["notification"]["title"] = title

            if self.sound:
                payload["notification"]["sound"] = self.sound

            if self.badge is not None:
                payload["notification"]["badge"] = self.badge

            self.logger.debug(
                "Pushy POST URL:"
                f" {notify_url} (cert_verify={self.verify_certificate!r})"
            )
            self.logger.debug(f"Pushy Payload: {payload!s}")

            # Always call throttle before any remote server i/o is made
            self.throttle()

            try:
                r = requests.post(
                    notify_url,
                    data=dumps(payload),
                    headers=headers,
                    verify=self.verify_certificate,
                    timeout=self.request_timeout,
                )

                # Sample response
                # See: https://pushy.me/docs/api/send-notifications
                # {
                #     "success": true,
                #     "id": "5ea9b214b47cad768a35f13a",
                #     "info": {
                #         "devices": 1
                #         "failed": ['abc']
                #     }
                # }
                try:
                    content = loads(r.content)

                except (AttributeError, TypeError, ValueError):
                    # ValueError = r.content is Unparsable
                    # TypeError = r.content is None
                    # AttributeError = r is None
                    content = {
                        "success": False,
                        "id": "",
                        "info": {},
                    }

                if r.status_code != requests.codes.ok or not content.get(
                    "success"
                ):

                    # We had a problem
                    status_str = NotifyPushy.http_response_code_lookup(
                        r.status_code, PUSHY_HTTP_ERROR_MAP
                    )

                    self.logger.warning(
                        "Failed to send Pushy notification to {}: "
                        "{}{}error={}.".format(
                            target,
                            status_str,
                            ", " if status_str else "",
                            r.status_code,
                        )
                    )

                    self.logger.debug(f"Response Details:\r\n{r.content}")

                    has_error = True
                    continue

                else:
                    self.logger.info(f"Sent Pushy notification to {target}.")

            except requests.RequestException as e:
                self.logger.warning(
                    "A Connection error occurred sending Pushy:%s "
                    "notification",
                    target,
                )
                self.logger.debug(f"Socket Exception: {e!s}")

                has_error = True
                continue

        return not has_error

    @property
    def url_identifier(self):
        """Returns all of the identifiers that make this URL unique from
        another simliar one.

        Targets or end points should never be identified here.
        """
        return (self.secure_protocol, self.apikey)

    def url(self, privacy=False, *args, **kwargs):
        """Returns the URL built dynamically based on specified arguments."""

        # Define any URL parameters
        params = {}
        if self.sound:
            params["sound"] = self.sound

        if self.badge is not None:
            params["badge"] = str(self.badge)

        # Extend our parameters
        params.update(self.url_parameters(privacy=privacy, *args, **kwargs))

        return "{schema}://{apikey}/{targets}/?{params}".format(
            schema=self.secure_protocol,
            apikey=self.pprint(self.apikey, privacy, safe=""),
            targets="/".join([
                NotifyPushy.quote(x, safe="@#")
                for x in chain(
                    # Topics are prefixed with a pound/hashtag symbol
                    [f"#{x}" for x in self.topics],
                    # Devices
                    [f"@{x}" for x in self.devices],
                )
            ]),
            params=NotifyPushy.urlencode(params),
        )

    def __len__(self):
        """Returns the number of targets associated with this notification."""
        return len(self.topics) + len(self.devices)

    @staticmethod
    def parse_url(url):
        """Parses the URL and returns enough arguments that can allow us to re-
        instantiate this object."""
        results = NotifyBase.parse_url(url, verify_host=False)
        if not results:
            # We're done early as we couldn't load the results
            return results

        # Token
        results["apikey"] = NotifyPushy.unquote(results["host"])

        # Retrieve all of our targets
        results["targets"] = NotifyPushy.split_path(results["fullpath"])

        # Get the sound
        if "sound" in results["qsd"] and len(results["qsd"]["sound"]):
            results["sound"] = NotifyPushy.unquote(results["qsd"]["sound"])

        # Badge
        if "badge" in results["qsd"] and results["qsd"]["badge"]:
            results["badge"] = NotifyPushy.unquote(
                results["qsd"]["badge"].strip()
            )

        # Support key variable to store Secret API Key
        if "key" in results["qsd"] and len(results["qsd"]["key"]):
            results["apikey"] = results["qsd"]["key"]

        # The 'to' makes it easier to use yaml configuration
        if "to" in results["qsd"] and len(results["qsd"]["to"]):
            results["targets"] += NotifyPushy.parse_list(results["qsd"]["to"])

        return results



# BSD 2-Clause License
#
# Apprise - Push Notification Library.
# Copyright (c) 2025, Chris Caron <lead2gold@gmail.com>
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice,
#    this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

# Sign-up with https://burstsms.com/
#
# Define your API Secret here and acquire your API Key
#  - https://can.transmitsms.com/profile
#
import requests

from ..common import NotifyType
from ..locale import gettext_lazy as _
from ..url import PrivacyMode
from ..utils.parse import (
    is_phone_no,
    parse_bool,
    parse_phone_no,
    validate_regex,
)
from .base import NotifyBase


class BurstSMSCountryCode:
    # Australia
    AU = "au"
    # New Zeland
    NZ = "nz"
    # United Kingdom
    UK = "gb"
    # United States
    US = "us"


BURST_SMS_COUNTRY_CODES = (
    BurstSMSCountryCode.AU,
    BurstSMSCountryCode.NZ,
    BurstSMSCountryCode.UK,
    BurstSMSCountryCode.US,
)


class NotifyBurstSMS(NotifyBase):
    """A wrapper for Burst SMS Notifications."""

    # The default descriptive name associated with the Notification
    service_name = "Burst SMS"

    # The services URL
    service_url = "https://burstsms.com/"

    # The default protocol
    secure_protocol = "burstsms"

    # The maximum amount of SMS Messages that can reside within a single
    # batch transfer based on:
    #  https://developer.transmitsms.com/#74911cf8-dec6-4319-a499-7f535a7fd08c
    default_batch_size = 500

    # A URL that takes you to the setup/help of the specific protocol
    setup_url = "https://github.com/caronc/apprise/wiki/Notify_burst_sms"

    # Burst SMS uses the http protocol with JSON requests
    notify_url = "https://api.transmitsms.com/send-sms.json"

    # The maximum length of the body
    body_maxlen = 160

    # A title can not be used for SMS Messages.  Setting this to zero will
    # cause any title (if defined) to get placed into the message body.
    title_maxlen = 0

    # Define object templates
    templates = ("{schema}://{apikey}:{secret}@{sender_id}/{targets}",)

    # Define our template tokens
    template_tokens = dict(
        NotifyBase.template_tokens,
        **{
            "apikey": {
                "name": _("API Key"),
                "type": "string",
                "required": True,
                "regex": (r"^[a-z0-9]+$", "i"),
                "private": True,
            },
            "secret": {
                "name": _("API Secret"),
                "type": "string",
                "private": True,
                "required": True,
                "regex": (r"^[a-z0-9]+$", "i"),
            },
            "sender_id": {
                "name": _("Sender ID"),
                "type": "string",
                "required": True,
                "map_to": "source",
            },
            "target_phone": {
                "name": _("Target Phone No"),
                "type": "string",
                "prefix": "+",
                "regex": (r"^[0-9\s)(+-]+$", "i"),
                "map_to": "targets",
            },
            "targets": {
                "name": _("Targets"),
                "type": "list:string",
                "required": True,
            },
        },
    )

    # Define our template arguments
    template_args = dict(
        NotifyBase.template_args,
        **{
            "to": {
                "alias_of": "targets",
            },
            "from": {
                "alias_of": "sender_id",
            },
            "key": {
                "alias_of": "apikey",
            },
            "secret": {
                "alias_of": "secret",
            },
            "country": {
                "name": _("Country"),
                "type": "choice:string",
                "values": BURST_SMS_COUNTRY_CODES,
                "default": BurstSMSCountryCode.US,
            },
            # Validity
            # Expire a message send if it is undeliverable (defined in minutes)
            # If set to Zero (0); this is the default and sets the max validity
            # period
            "validity": {"name": _("validity"), "type": "int", "default": 0},
            "batch": {
                "name": _("Batch Mode"),
                "type": "bool",
                "default": False,
            },
        },
    )

    def __init__(
        self,
        apikey,
        secret,
        source,
        targets=None,
        country=None,
        validity=None,
        batch=None,
        **kwargs,
    ):
        """Initialize Burst SMS Object."""
        super().__init__(**kwargs)

        # API Key (associated with project)
        self.apikey = validate_regex(
            apikey, *self.template_tokens["apikey"]["regex"]
        )
        if not self.apikey:
            msg = f"An invalid Burst SMS API Key ({apikey}) was specified."
            self.logger.warning(msg)
            raise TypeError(msg)

        # API Secret (associated with project)
        self.secret = validate_regex(
            secret, *self.template_tokens["secret"]["regex"]
        )
        if not self.secret:
            msg = f"An invalid Burst SMS API Secret ({secret}) was specified."
            self.logger.warning(msg)
            raise TypeError(msg)

        if not country:
            self.country = self.template_args["country"]["default"]

        else:
            self.country = country.lower().strip()
            if country not in BURST_SMS_COUNTRY_CODES:
                msg = (
                    f"An invalid Burst SMS country ({country}) was specified."
                )
                self.logger.warning(msg)
                raise TypeError(msg)

        # Set our Validity
        self.validity = self.template_args["validity"]["default"]
        if validity:
            try:
                self.validity = int(validity)

            except (ValueError, TypeError):
                msg = (
                    f"The Burst SMS Validity specified ({validity}) is"
                    " invalid."
                )
                self.logger.warning(msg)
                raise TypeError(msg) from None

        # Prepare Batch Mode Flag
        self.batch = (
            self.template_args["batch"]["default"] if batch is None else batch
        )

        # The Sender ID
        self.source = validate_regex(source)
        if not self.source:
            msg = f"The Account Sender ID specified ({source}) is invalid."
            self.logger.warning(msg)
            raise TypeError(msg)

        # Parse our targets
        self.targets = []

        for target in parse_phone_no(targets):
            # Validate targets and drop bad ones:
            result = is_phone_no(target)
            if not result:
                self.logger.warning(
                    f"Dropped invalid phone # ({target}) specified.",
                )
                continue

            # store valid phone number
            self.targets.append(result["full"])

        return

    def send(self, body, title="", notify_type=NotifyType.INFO, **kwargs):
        """Perform Burst SMS Notification."""

        if not self.targets:
            self.logger.warning(
                "There are no valid Burst SMS targets to notify."
            )
            return False

        # error tracking (used for function return)
        has_error = False

        # Prepare our headers
        headers = {
            "User-Agent": self.app_id,
            "Accept": "application/json",
        }

        # Prepare our authentication
        auth = (self.apikey, self.secret)

        # Prepare our payload
        payload = {
            "countrycode": self.country,
            "message": body,
            # Sender ID
            "from": self.source,
            # The to gets populated in the loop below
            "to": None,
        }

        # Send in batches if identified to do so
        batch_size = 1 if not self.batch else self.default_batch_size

        # Create a copy of the targets list
        targets = list(self.targets)

        for index in range(0, len(targets), batch_size):

            # Prepare our user
            payload["to"] = ",".join(self.targets[index : index + batch_size])

            # Some Debug Logging
            self.logger.debug(
                "Burst SMS POST URL:"
                f" {self.notify_url} (cert_verify={self.verify_certificate})"
            )
            self.logger.debug(f"Burst SMS Payload: {payload}")

            # Always call throttle before any remote server i/o is made
            self.throttle()

            try:
                r = requests.post(
                    self.notify_url,
                    data=payload,
                    headers=headers,
                    auth=auth,
                    verify=self.verify_certificate,
                    timeout=self.request_timeout,
                )

                if r.status_code != requests.codes.ok:
                    # We had a problem
                    status_str = NotifyBurstSMS.http_response_code_lookup(
                        r.status_code
                    )

                    self.logger.warning(
                        "Failed to send Burst SMS notification to {} "
                        "target(s): {}{}error={}.".format(
                            len(self.targets[index : index + batch_size]),
                            status_str,
                            ", " if status_str else "",
                            r.status_code,
                        )
                    )

                    self.logger.debug(f"Response Details:\r\n{r.content}")

                    # Mark our failure
                    has_error = True
                    continue

                else:
                    self.logger.info(
                        "Sent Burst SMS notification to "
                        f"{len(self.targets[index : index + batch_size])} "
                        "target(s)."
                    )

            except requests.RequestException as e:
                self.logger.warning(
                    f"A Connection error occurred sending Burst SMS "
                    "notification to "
                    f"{len(self.targets[index : index + batch_size])} "
                    "target(s)."
                )
                self.logger.debug(f"Socket Exception: {e!s}")

                # Mark our failure
                has_error = True
                continue

        return not has_error

    def url(self, privacy=False, *args, **kwargs):
        """Returns the URL built dynamically based on specified arguments."""

        # Define any URL parameters
        params = {
            "country": self.country,
            "batch": "yes" if self.batch else "no",
        }

        if self.validity:
            params["validity"] = str(self.validity)

        # Extend our parameters
        params.update(self.url_parameters(privacy=privacy, *args, **kwargs))

        return "{schema}://{key}:{secret}@{source}/{targets}/?{params}".format(
            schema=self.secure_protocol,
            key=self.pprint(self.apikey, privacy, safe=""),
            secret=self.pprint(
                self.secret, privacy, mode=PrivacyMode.Secret, safe=""
            ),
            source=NotifyBurstSMS.quote(self.source, safe=""),
            targets="/".join(
                [NotifyBurstSMS.quote(x, safe="") for x in self.targets]
            ),
            params=NotifyBurstSMS.urlencode(params),
        )

    @property
    def url_identifier(self):
        """Returns all of the identifiers that make this URL unique from
        another simliar one.

        Targets or end points should never be identified here.
        """
        return (self.secure_protocol, self.apikey, self.secret, self.source)

    def __len__(self):
        """Returns the number of targets associated with this notification."""
        #
        # Factor batch into calculation
        #
        batch_size = 1 if not self.batch else self.default_batch_size
        targets = len(self.targets)
        if batch_size > 1:
            targets = int(targets / batch_size) + (
                1 if targets % batch_size else 0
            )

        return targets if targets > 0 else 1

    @staticmethod
    def parse_url(url):
        """Parses the URL and returns enough arguments that can allow us to re-
        instantiate this object."""
        results = NotifyBase.parse_url(url, verify_host=False)
        if not results:
            # We're done early as we couldn't load the results
            return results

        # The hostname is our source (Sender ID)
        results["source"] = NotifyBurstSMS.unquote(results["host"])

        # Get any remaining targets
        results["targets"] = NotifyBurstSMS.split_path(results["fullpath"])

        # Get our account_side and auth_token from the user/pass config
        results["apikey"] = NotifyBurstSMS.unquote(results["user"])
        results["secret"] = NotifyBurstSMS.unquote(results["password"])

        # API Key
        if "key" in results["qsd"] and len(results["qsd"]["key"]):
            # Extract the API Key from an argument
            results["apikey"] = NotifyBurstSMS.unquote(results["qsd"]["key"])

        # API Secret
        if "secret" in results["qsd"] and len(results["qsd"]["secret"]):
            # Extract the API Secret from an argument
            results["secret"] = NotifyBurstSMS.unquote(
                results["qsd"]["secret"]
            )

        # Support the 'from'  and 'source' variable so that we can support
        # targets this way too.
        # The 'from' makes it easier to use yaml configuration
        if "from" in results["qsd"] and len(results["qsd"]["from"]):
            results["source"] = NotifyBurstSMS.unquote(results["qsd"]["from"])
        if "source" in results["qsd"] and len(results["qsd"]["source"]):
            results["source"] = NotifyBurstSMS.unquote(
                results["qsd"]["source"]
            )

        # Support country
        if "country" in results["qsd"] and len(results["qsd"]["country"]):
            results["country"] = NotifyBurstSMS.unquote(
                results["qsd"]["country"]
            )

        # Support validity value
        if "validity" in results["qsd"] and len(results["qsd"]["validity"]):
            results["validity"] = NotifyBurstSMS.unquote(
                results["qsd"]["validity"]
            )

        # Get Batch Mode Flag
        if "batch" in results["qsd"] and len(results["qsd"]["batch"]):
            results["batch"] = parse_bool(results["qsd"]["batch"])

        # Support the 'to' variable so that we can support rooms this way too
        # The 'to' makes it easier to use yaml configuration
        if "to" in results["qsd"] and len(results["qsd"]["to"]):
            results["targets"] += NotifyBurstSMS.parse_phone_no(
                results["qsd"]["to"]
            )

        return results



#
# Copyright (C) 2025 Chris Caron <lead2gold@gmail.com>
# All rights reserved.
#
# This code is licensed under the MIT License.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files(the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and / or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions :
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
import base64
import binascii
import builtins
import contextlib
from datetime import datetime, timedelta, timezone
import glob
import gzip
import hashlib
from itertools import chain
import json
import os
import re
import tempfile
import time
from typing import Any, Optional, Union
import zlib

from . import exception
from .common import PersistentStoreMode
from .logger import logger
from .utils.disk import path_decode

# Used for writing/reading time stored in cache file
EPOCH = datetime(1970, 1, 1, tzinfo=timezone.utc)

# isoformat is spelled out for compatibility with Python v3.6
AWARE_DATE_ISO_FORMAT = "%Y-%m-%dT%H:%M:%S.%f%z"
NAIVE_DATE_ISO_FORMAT = "%Y-%m-%dT%H:%M:%S.%f"


def _ntf_tidy(ntf):
    """Reusable NamedTemporaryFile Cleanup."""
    if ntf:
        # Cleanup
        with contextlib.suppress(OSError):
            ntf.close()

        try:
            os.unlink(ntf.name)
            logger.trace("Persistent temporary file removed: %s", ntf.name)

        except (FileNotFoundError, AttributeError):
            # AttributeError: something weird was passed in, no action required
            # FileNotFound: no worries; we were removing it anyway
            pass

        except OSError as e:
            logger.error(
                "Persistent temporary file removal failed: %s", ntf.name
            )
            logger.debug("Persistent Storage Exception: %s", str(e))


class CacheObject:

    hash_engine = hashlib.sha256
    hash_length = 6

    def __init__(
        self,
        value: Any = None,
        expires: Union[bool, float, int, datetime, None] = False,
        persistent: bool = True,
    ) -> None:
        """Tracks our objects and associates a time limit with them."""

        self.__value = value
        self.__class_name = value.__class__.__name__
        self.__expires = None

        if expires:
            self.set_expiry(expires)

        # Whether or not we persist this object to disk or not
        self.__persistent = bool(persistent)

    def set(
        self,
        value: Any,
        expires: Union[bool, float, int, datetime, None] = None,
        persistent: Optional[bool] = None,
    ) -> None:
        """Sets fields on demand, if set to none, then they are left as is.

        The intent of set is that it allows you to set a new a value and
        optionally alter meta information against it.

        If expires or persistent isn't specified then their previous values are
        used.
        """

        self.__value = value
        self.__class_name = value.__class__.__name__
        if expires is not None:
            self.set_expiry(expires)

        if persistent is not None:
            self.__persistent = bool(persistent)

    def set_expiry(self, expires:
                   Union[datetime, bool, float, int, None] = None) -> None:
        """Sets a new expiry."""

        if isinstance(expires, datetime):
            self.__expires = expires.astimezone(timezone.utc)

        elif expires in (None, False):
            # Accepted - no expiry
            self.__expires = None

        elif expires is True:
            # Force expiry to now
            self.__expires = datetime.now(tz=timezone.utc)

        elif isinstance(expires, (float, int)):
            self.__expires = datetime.now(tz=timezone.utc) + timedelta(
                seconds=expires
            )

        else:  # Unsupported
            raise AttributeError(
                f"An invalid expiry time ({expires} was specified"
            )

    def hash(self) -> str:
        """Our checksum to track the validity of our data."""
        return self.hash_engine(
            str(self).encode("utf-8"), usedforsecurity=False
        ).hexdigest()

    def json(self) -> Optional[dict[str, Any]]:
        """Returns our preparable json object."""

        return {
            "v": self.__value,
            "x": (
                (self.__expires - EPOCH).total_seconds()
                if self.__expires
                else None
            ),
            "c": (
                self.__class_name
                if not isinstance(self.__value, datetime)
                else (
                    "aware_datetime"
                    if self.__value.tzinfo
                    else "naive_datetime"
                )
            ),
            "!": self.hash()[: self.hash_length],
        }

    @staticmethod
    def instantiate(
        content: dict[str, Any],
        persistent: bool = True,
        verify: bool = True,
    ) -> Optional["CacheObject"]:
        """Loads back data read in and returns a CacheObject or None if it
        could not be loaded.

        You can pass in the contents of CacheObject.json() and you'll receive a
        copy assuming the hash checks okay
        """
        try:
            value = content["v"]
            expires = content["x"]
            if expires is not None:
                expires = datetime.fromtimestamp(expires, timezone.utc)

            # Acquire some useful integrity objects
            class_name = content.get("c", "")
            if not isinstance(class_name, str):
                raise TypeError("Class name not expected string")

            hashsum = content.get("!", "")
            if not isinstance(hashsum, str):
                raise TypeError("SHA1SUM not expected string")

        except (TypeError, KeyError) as e:
            logger.trace(f"CacheObject could not be parsed from {content}")
            logger.trace("CacheObject exception: %s", str(e))
            return None

        if class_name in ("aware_datetime", "naive_datetime", "datetime"):
            # If datetime is detected, it will fall under the naive category
            iso_format = (
                AWARE_DATE_ISO_FORMAT
                if class_name[0] == "a"
                else NAIVE_DATE_ISO_FORMAT
            )
            try:
                # Python v3.6 Support
                value = datetime.strptime(value, iso_format)

            except (TypeError, ValueError):
                # TypeError is thrown if content is not string
                # ValueError is thrown if the string is not a valid format
                logger.trace(
                    f"CacheObject (dt) corrupted loading from {content}"
                )
                return None

        elif class_name == "bytes":
            try:
                # Convert our object back to a bytes
                value = base64.b64decode(value)

            except binascii.Error:
                logger.trace(
                    f"CacheObject (bin) corrupted loading from {content}"
                )
                return None

        # Initialize our object
        co = CacheObject(value, expires, persistent=persistent)
        if verify and co.hash()[: co.hash_length] != hashsum:
            # Our object was tampered with
            logger.debug(f"Tampering detected with cache entry {co}")
            del co
            return None

        return co

    @property
    def value(self) -> Any:
        """Returns our value."""
        return self.__value

    @property
    def persistent(self) -> bool:
        """Returns our persistent value."""
        return self.__persistent

    @property
    def expires(self) -> Optional[datetime]:
        """Returns the datetime the object will expire."""
        return self.__expires

    @property
    def expires_sec(self) -> Optional[float]:
        """Returns the number of seconds from now the object will expire."""

        return (
            None
            if self.__expires is None
            else max(
                0.0,
                (
                    self.__expires - datetime.now(tz=timezone.utc)
                ).total_seconds(),
            )
        )

    def __bool__(self) -> bool:
        """Returns True it the object hasn't expired, and False if it has."""
        if self.__expires is None:
            # No Expiry
            return True

        # Calculate if we've expired or not
        return self.__expires > datetime.now(tz=timezone.utc)

    def __eq__(self, other) -> bool:
        """Handles equality == flag."""
        if isinstance(other, CacheObject):
            return str(self) == str(other)

        return self.__value == other

    def __str__(self) -> str:
        """String output of our data."""
        persistent = "+" if self.persistent else "-"
        return f"{self.__class_name}:{persistent}:{self.__value} expires: " + (
            "never"
            if self.__expires is None
            else self.__expires.strftime(NAIVE_DATE_ISO_FORMAT)
        )


class CacheJSONEncoder(json.JSONEncoder):
    """A JSON Encoder for handling each of our cache objects."""

    def default(self, entry):
        if isinstance(entry, datetime):
            return entry.strftime(
                AWARE_DATE_ISO_FORMAT
                if entry.tzinfo is not None
                else NAIVE_DATE_ISO_FORMAT
            )

        elif isinstance(entry, CacheObject):
            return entry.json()

        elif isinstance(entry, bytes):
            return base64.b64encode(entry).decode("utf-8")

        return super().default(entry)


class PersistentStore:
    """An object to make working with persistent storage easier.

    read() and write() are used for direct file i/o

    set(), get() are used for caching
    """

    # The maximum file-size we will allow the persistent store to grow to
    # 1 MB = 1048576 bytes
    max_file_size = 1048576

    # 30 days in seconds
    default_file_expiry = 2678400

    # File encoding to use
    encoding = "utf-8"

    # Default data set
    base_key = "default"

    # Directory to store cache
    __cache_key = "cache"

    # Our Temporary working directory
    temp_dir = "tmp"

    # The directory our persistent store content gets placed in
    data_dir = "var"

    # Our Persistent Store File Extension
    __extension = ".psdata"

    # Identify our backup file extension
    __backup_extension = "._psbak"

    # Used to verify the key specified is valid
    #  - must start with an alpha_numeric
    #  - following optional characters can include period, underscore and
    #    equal
    __valid_key = re.compile(r"[a-z0-9][a-z0-9._-]*", re.I)

    # Reference only
    __not_found_ref = (None, None)

    def __init__(
        self,
        path: Optional[str] = None,
        namespace: str = "default",
        mode: Optional[Union[str, PersistentStoreMode]] = None,
    ) -> None:
        """Provide the namespace to work within.

        namespaces can only contain alpha-numeric characters with the exception
        of '-' (dash), '_' (underscore), and '.' (period). The namespace must
        be be relative to the current URL being controlled.
        """
        # Initalize our mode so __del__() calls don't go bad on the
        # error checking below
        self.__mode = None

        # Populated only once and after size() is called
        self.__exclude_list = None

        # Files to renew on calls to flush
        self.__renew = set()

        if not isinstance(namespace, str) or not self.__valid_key.match(
            namespace
        ):
            raise AttributeError(
                f"Persistent Storage namespace ({namespace}) provided is"
                " invalid"
            )

        if isinstance(path, str):
            # A storage path has been defined
            if mode is None:
                # Store Default if no mode was provided along side of it
                mode = PersistentStoreMode.AUTO

            # Store our information
            self.__base_path = os.path.join(path_decode(path), namespace)
            self.__temp_path = os.path.join(self.__base_path, self.temp_dir)
            self.__data_path = os.path.join(self.__base_path, self.data_dir)

        else:  # If no storage path is provide we set our mode to MEMORY
            mode = PersistentStoreMode.MEMORY
            self.__base_path = None
            self.__temp_path = None
            self.__data_path = None

        # Tracks when we have content to flush
        self.__dirty = False

        # A caching value to track persistent storage disk size
        self.__cache_size = None
        self.__cache_files = {}

        # Internal Cache
        self._cache = None

        try:
            # Store our mode
            self.__mode = (
                mode if isinstance(mode, PersistentStoreMode)
                else PersistentStoreMode(mode.lower())
            )

        except (AttributeError, ValueError):
            err = (
                f"An invalid persistent storage mode ({mode}) was specified.",
            )
            logger.warning(err)
            raise AttributeError(err) from None

        # Prepare our environment
        self.__prepare()

    def read(
        self,
        key: Optional[str] = None,
        compress: bool = True,
        expires: Union[bool, float, int] = False,
    ) -> Optional[bytes]:
        """Returns the content of the persistent store object.

        if refresh is set to True, then the file's modify time is updated
        preventing it from getting caught in prune calls.  It's a means of
        allowing it to persist and not get cleaned up in later prune calls.

        Content is always returned as a byte object
        """
        try:
            with self.open(key, mode="rb", compress=compress) as fd:
                results = fd.read(self.max_file_size)
                if expires is False:
                    self.__renew.add(
                        os.path.join(
                            self.__data_path, f"{key}{self.__extension}"
                        )
                    )

                return results

        except (FileNotFoundError, exception.AppriseDiskIOError):
            # FileNotFoundError: No problem
            # exception.AppriseDiskIOError:
            #   - Logging of error already occurred inside self.open()
            pass

        except (OSError, zlib.error, EOFError, UnicodeDecodeError) as e:
            # We can't access the file or it does not exist
            logger.warning("Could not read with persistent key: %s", key)
            logger.debug("Persistent Storage Exception: %s", str(e))

        # return none
        return None

    def write(
        self,
        data: Union[bytes, str, Any],
        key: Optional[str] = None,
        compress: bool = True,
        _recovery: bool = False,
    ) -> bool:
        """Writes the content to the persistent store if it doesn't exceed our
        filesize limit.

        Content is always written as a byte object

        _recovery is reserved for internal usage and should not be changed
        """

        if key is None:
            key = self.base_key

        elif not isinstance(key, str) or not self.__valid_key.match(key):
            raise AttributeError(
                f"Persistent Storage key ({key} provided is invalid"
            )

        if not isinstance(data, (bytes, str)):
            # One last check, we will accept read() objets with the expectation
            # it will return a binary dataset
            if not (hasattr(data, "read") and callable(data.read)):
                raise AttributeError(
                    f"Invalid data type {type(data)} provided to Persistent"
                    " Storage"
                )

            try:
                # Read in our data
                data = data.read()
                if not isinstance(data, (bytes, str)):
                    raise AttributeError(
                        f"Invalid data type {type(data)} provided to"
                        " Persistent Storage"
                    )

            except Exception as e:
                logger.warning(
                    "Could read() from potential iostream with persistent "
                    "key: %s",
                    key,
                )
                logger.debug("Persistent Storage Exception: %s", str(e))
                raise exception.AppriseDiskIOError(
                    f"Invalid data type {type(data)} provided to Persistent"
                    " Storage"
                ) from None

        if self.__mode == PersistentStoreMode.MEMORY:
            # Nothing further can be done
            return False

        if _recovery:
            # Attempt to recover from a bad directory structure or setup
            self.__prepare()

        # generate our filename based on the key provided
        io_file = os.path.join(self.__data_path, f"{key}{self.__extension}")

        # Calculate the files current filesize
        try:
            prev_size = os.stat(io_file).st_size

        except FileNotFoundError:
            # No worries, no size to accomodate
            prev_size = 0

        except OSError as e:
            # Permission error of some kind or disk problem...
            # There is nothing we can do at this point
            logger.warning("Could not write with persistent key: %s", key)
            logger.debug("Persistent Storage Exception: %s", str(e))
            return False

        # Create a temporary file to write our content into
        # ntf = NamedTemporaryFile
        ntf = None
        new_file_size = 0
        try:
            if isinstance(data, str):
                data = data.encode(self.encoding)

            ntf = tempfile.NamedTemporaryFile(  # noqa: SIM115
                mode="wb", dir=self.__temp_path, delete=False
            )

            # Close our file
            ntf.close()

            # Pointer to our open call
            _open = open if not compress else gzip.open

            with _open(ntf.name, mode="wb") as fd:
                # Write our content
                fd.write(data)

            # Get our file size
            new_file_size = os.stat(ntf.name).st_size

            # Log our progress
            logger.trace(
                "Wrote %d bytes of data to persistent key: %s",
                new_file_size,
                key,
            )

        except FileNotFoundError:
            # This happens if the directory path is gone preventing the file
            # from being created...
            if not _recovery:
                return self.write(
                    data=data, key=key, compress=compress, _recovery=True
                )

            # We've already made our best effort to recover if we are here in
            # our code base... we're going to have to exit

            # Tidy our Named Temporary File
            _ntf_tidy(ntf)

            # Early Exit
            return False

        except (OSError, UnicodeEncodeError, zlib.error) as e:
            # We can't access the file or it does not exist
            logger.warning("Could not write to persistent key: %s", key)
            logger.debug("Persistent Storage Exception: %s", str(e))

            # Tidy our Named Temporary File
            _ntf_tidy(ntf)

            return False

        if (
            self.max_file_size > 0
            and (new_file_size + self.size() - prev_size) > self.max_file_size
        ):
            # The content to store is to large
            logger.warning(
                "Persistent content exceeds allowable maximum file length"
                f" ({int(self.max_file_size / 1024)}KB); provide"
                f" {int(new_file_size / 1024)}KB"
            )
            return False

        # Return our final move
        if not self.__move(ntf.name, io_file):
            # Attempt to restore things as they were

            # Tidy our Named Temporary File
            _ntf_tidy(ntf)
            return False

        # Resetour reference variables
        self.__cache_size = None
        self.__cache_files.clear()

        # Content installed
        return True

    def __move(self, src, dst):
        """Moves the new file in place and handles the old if it exists already
        If the transaction fails in any way, the old file is swapped back.

        Function returns True if successful and False if not.
        """

        # A temporary backup of the file we want to move in place
        dst_backup = (
            dst[: -len(self.__backup_extension)] + self.__backup_extension
        )

        #
        # Backup the old file (if it exists) allowing us to have a restore
        # point in the event of a failure
        #
        try:
            # make sure the file isn't already present; if it is; remove it
            os.unlink(dst_backup)
            logger.trace(
                "Removed previous persistent backup file: %s", dst_backup
            )

        except FileNotFoundError:
            # no worries; we were removing it anyway
            pass

        except OSError as e:
            # Permission error of some kind or disk problem...
            # There is nothing we can do at this point
            logger.warning(
                "Could not previous persistent data backup: %s", dst_backup
            )
            logger.debug("Persistent Storage Exception: %s", str(e))
            return False

        try:
            # Back our file up so we have a fallback
            os.rename(dst, dst_backup)
            logger.trace(
                "Persistent storage backup file created: %s", dst_backup
            )

        except FileNotFoundError:
            # Not a problem; this is a brand new file we're writing
            # There is nothing to backup
            pass

        except OSError as e:
            # This isn't good... we couldn't put our new file in place
            logger.warning(
                "Could not install persistent content %s -> %s",
                dst,
                os.path.basename(dst_backup),
            )
            logger.debug("Persistent Storage Exception: %s", str(e))
            return False

        #
        # Now place the new file
        #
        try:
            os.rename(src, dst)
            logger.trace("Persistent file installed: %s", dst)

        except OSError as e:
            # This isn't good... we couldn't put our new file in place
            # Begin fall-back process before leaving the funtion
            logger.warning(
                "Could not install persistent content %s -> %s",
                src,
                os.path.basename(dst),
            )
            logger.debug("Persistent Storage Exception: %s", str(e))
            try:
                # Restore our old backup (if it exists)
                os.rename(dst_backup, dst)
                logger.trace("Restoring original persistent content: %s", dst)

            except FileNotFoundError:
                # Not a problem
                pass

            except OSError as e:
                # Permission error of some kind or disk problem...
                # There is nothing we can do at this point
                logger.warning(
                    "Failed to restore original persistent file: %s", dst
                )
                logger.debug("Persistent Storage Exception: %s", str(e))

            return False

        return True

    def open(
        self,
        key: Optional[str] = None,
        mode: str = "r",
        buffering: int = -1,
        encoding: Optional[str] = None,
        errors: Optional[str] = None,
        newline: Optional[str] = None,
        closefd: bool = True,
        opener: Optional[Any] = None,
        compress: bool = False,
        compresslevel: int = 9,
    ) -> Any:
        """Returns an iterator to our our file within our namespace identified
        by the key provided.

        If no key is provided, then the default is used
        """

        if key is None:
            key = self.base_key

        elif not isinstance(key, str) or not self.__valid_key.match(key):
            raise AttributeError(
                f"Persistent Storage key ({key} provided is invalid"
            )

        if self.__mode == PersistentStoreMode.MEMORY:
            # Nothing further can be done
            raise FileNotFoundError()

        io_file = os.path.join(self.__data_path, f"{key}{self.__extension}")
        try:
            return (
                open(
                    io_file,
                    mode=mode,
                    buffering=buffering,
                    encoding=encoding,
                    errors=errors,
                    newline=newline,
                    closefd=closefd,
                    opener=opener,
                )
                if not compress
                else gzip.open(
                    io_file,
                    compresslevel=compresslevel,
                    encoding=encoding,
                    errors=errors,
                    newline=newline,
                )
            )

        except FileNotFoundError:
            # pass along (but wrap with Apprise exception)
            raise exception.AppriseFileNotFound(
                f"No such file or directory: '{io_file}'"
            ) from None

        except (OSError, zlib.error) as e:
            # We can't access the file or it does not exist
            logger.warning("Could not read with persistent key: %s", key)
            logger.debug("Persistent Storage Exception: %s", str(e))
            raise exception.AppriseDiskIOError(str(e)) from None

    def get(
        self,
        key: str,
        default: Any = None,
        lazy: bool = True,
    ) -> Any:
        """Fetches from cache."""

        if self._cache is None and not self.__load_cache():
            return default

        if (
            key in self._cache
            and self.__mode != PersistentStoreMode.MEMORY
            and not self.__dirty
        ):

            # ensure we renew our content
            self.__renew.add(self.cache_file)

        return self._cache[key].value if self._cache.get(key) else default

    def set(
        self,
        key: str,
        value: Any,
        expires: Union[float, int, datetime, bool, None] = None,
        persistent: bool = True,
        lazy: bool = True,
    ) -> bool:
        """Cache reference."""

        if self._cache is None and not self.__load_cache():
            return False

        cache = CacheObject(value, expires, persistent=persistent)
        # Fetch our cache value
        try:
            if lazy and cache == self._cache[key]:
                # We're done; nothing further to do
                return True

        except KeyError:
            pass

        # Store our new cache
        self._cache[key] = CacheObject(value, expires, persistent=persistent)

        # Set our dirty flag
        self.__dirty = persistent

        if self.__dirty and self.__mode == PersistentStoreMode.FLUSH:
            # Flush changes to disk
            return self.flush()

        return True

    def clear(self, *args: str) -> Optional[bool]:
        """Remove one or more cache entry by it's key.

            e.g: clear('key')
                 clear('key1', 'key2', key-12')

        Or clear everything:
                 clear()
        """
        if self._cache is None and not self.__load_cache():
            return False

        if args:
            for arg in args:

                try:
                    del self._cache[arg]

                    # Set our dirty flag (if not set already)
                    self.__dirty = True

                except KeyError:
                    pass

        elif self._cache:
            # Request to remove everything and there is something to remove

            # Set our dirty flag (if not set already)
            self.__dirty = True

            # Reset our object
            self._cache.clear()

        if self.__dirty and self.__mode == PersistentStoreMode.FLUSH:
            # Flush changes to disk
            return self.flush()

    def prune(self) -> bool:
        """Eliminates expired cache entries."""
        if self._cache is None and not self.__load_cache():
            return False

        change = False
        for key in list(self._cache.keys()):
            if key not in self:
                # It's identified as being expired
                if not change and self._cache[key].persistent:
                    # track change only if content was persistent
                    change = True

                    # Set our dirty flag
                    self.__dirty = True

                del self._cache[key]

        if self.__dirty and self.__mode == PersistentStoreMode.FLUSH:
            # Flush changes to disk
            return self.flush()

        return change

    def __load_cache(self, _recovery=False):
        """Loads our cache.

        _recovery is reserved for internal usage and should not be changed
        """

        # Prepare our dirty flag
        self.__dirty = False

        if self.__mode == PersistentStoreMode.MEMORY:
            # Nothing further to do
            self._cache = {}
            return True

        # Prepare our cache file
        cache_file = self.cache_file
        try:
            with gzip.open(cache_file, "rb") as f:
                # Read our ontent from disk
                self._cache = {}
                for k, v in json.loads(f.read().decode(self.encoding)).items():
                    co = CacheObject.instantiate(v)
                    if co:
                        # Verify our object before assigning it
                        self._cache[k] = co

                    elif not self.__dirty:
                        # Track changes from our loadset
                        self.__dirty = True

        except (
            UnicodeDecodeError,
            json.decoder.JSONDecodeError,
            zlib.error,
            TypeError,
            AttributeError,
            EOFError,
        ):

            # Let users known there was a problem
            logger.warning(
                "Corrupted access persistent cache content: %s", cache_file
            )

            if not _recovery:
                try:
                    os.unlink(cache_file)
                    logger.trace(
                        "Removed previous persistent cache content: %s",
                        cache_file,
                    )

                except FileNotFoundError:
                    # no worries; we were removing it anyway
                    pass

                except OSError as e:
                    # Permission error of some kind or disk problem...
                    # There is nothing we can do at this point
                    logger.warning(
                        "Could not remove persistent cache content: %s",
                        cache_file,
                    )
                    logger.debug("Persistent Storage Exception: %s", str(e))
                    return False
                return self.__load_cache(_recovery=True)

            return False

        except FileNotFoundError:
            # No problem; no cache to load
            self._cache = {}

        except OSError as e:
            # Permission error of some kind or disk problem...
            # There is nothing we can do at this point
            logger.warning(
                "Could not load persistent cache for namespace %s",
                os.path.basename(self.__base_path),
            )
            logger.debug("Persistent Storage Exception: %s", str(e))
            return False

        # Ensure our dirty flag is set to False
        return True

    def __prepare(self, flush=True):
        """Prepares a working environment."""
        if self.__mode != PersistentStoreMode.MEMORY:
            # Ensure our path exists
            try:
                os.makedirs(self.__base_path, mode=0o770, exist_ok=True)

            except OSError as e:
                # Permission error
                logger.debug(
                    "Could not create persistent store directory %s",
                    self.__base_path,
                )
                logger.debug("Persistent Storage Exception: %s", str(e))

                # Mode changed back to MEMORY
                self.__mode = PersistentStoreMode.MEMORY

            # Ensure our path exists
            try:
                os.makedirs(self.__temp_path, mode=0o770, exist_ok=True)

            except OSError as e:
                # Permission error
                logger.debug(
                    "Could not create persistent store directory %s",
                    self.__temp_path,
                )
                logger.debug("Persistent Storage Exception: %s", str(e))

                # Mode changed back to MEMORY
                self.__mode = PersistentStoreMode.MEMORY

            try:
                os.makedirs(self.__data_path, mode=0o770, exist_ok=True)

            except OSError as e:
                # Permission error
                logger.debug(
                    "Could not create persistent store directory %s",
                    self.__data_path,
                )
                logger.debug("Persistent Storage Exception: %s", str(e))

                # Mode changed back to MEMORY
                self.__mode = PersistentStoreMode.MEMORY

            if self.__mode is PersistentStoreMode.MEMORY:
                logger.warning(
                    "The persistent storage could not be fully initialized; "
                    "operating in MEMORY mode"
                )

            else:
                if self._cache:
                    # Recovery taking place
                    self.__dirty = True
                    logger.warning(
                        "The persistent storage environment was disrupted"
                    )

                    if self.__mode is PersistentStoreMode.FLUSH and flush:
                        # Flush changes to disk
                        return self.flush(_recovery=True)

    def flush(
        self,
        force: bool = False,
        _recovery: bool = False,
    ) -> bool:
        """Save's our cache to disk."""

        if self._cache is None or self.__mode == PersistentStoreMode.MEMORY:
            # nothing to do
            return True

        while self.__renew:
            # update our files
            path = self.__renew.pop()
            ftime = time.time()

            try:
                # (access_time, modify_time)
                os.utime(path, (ftime, ftime))
                logger.trace("file timestamp updated: %s", path)

            except FileNotFoundError:
                # No worries... move along
                pass

            except OSError as e:
                # We can't access the file or it does not exist
                logger.debug("Could not update file timestamp: %s", path)
                logger.debug("Persistent Storage Exception: %s", str(e))

        if not force and self.__dirty is False:
            # Nothing further to do
            logger.trace("Persistent cache is consistent with memory map")
            return True

        if _recovery:
            # Attempt to recover from a bad directory structure or setup
            self.__prepare(flush=False)

        # Unset our size lazy setting
        self.__cache_size = None
        self.__cache_files.clear()

        # Prepare our cache file
        cache_file = self.cache_file
        if not self._cache:
            #
            # We're deleting the cache file s there are no entries left in it
            #
            backup_file = (
                cache_file[: -len(self.__backup_extension)]
                + self.__backup_extension
            )

            try:
                os.unlink(backup_file)
                logger.trace(
                    "Removed previous persistent cache backup: %s", backup_file
                )

            except FileNotFoundError:
                # no worries; we were removing it anyway
                pass

            except OSError as e:
                # Permission error of some kind or disk problem...
                # There is nothing we can do at this point
                logger.warning(
                    "Could not remove persistent cache backup: %s", backup_file
                )
                logger.debug("Persistent Storage Exception: %s", str(e))
                return False

            try:
                os.rename(cache_file, backup_file)
                logger.trace(
                    "Persistent cache backup file created: %s", backup_file
                )

            except FileNotFoundError:
                # Not a problem; do not create a log entry
                pass

            except OSError as e:
                # This isn't good... we couldn't put our new file in place
                logger.warning(
                    "Could not remove stale persistent cache file: %s",
                    cache_file,
                )
                logger.debug("Persistent Storage Exception: %s", str(e))
                return False
            return True

        #
        # If we get here, we need to update our file based cache
        #

        # ntf = NamedTemporaryFile
        ntf = None

        try:
            ntf = tempfile.NamedTemporaryFile(  # noqa: SIM115
                mode="w+",
                encoding=self.encoding,
                dir=self.__temp_path,
                delete=False,
            )

            ntf.close()

        except FileNotFoundError:
            # This happens if the directory path is gone preventing the file
            # from being created...
            if not _recovery:
                return self.flush(force=True, _recovery=True)

            # We've already made our best effort to recover if we are here in
            # our code base... we're going to have to exit

            # Tidy our Named Temporary File
            _ntf_tidy(ntf)

            # Early Exit
            return False

        except OSError as e:
            logger.error(
                "Persistent temporary directory inaccessible: %s",
                self.__temp_path,
            )
            logger.debug("Persistent Storage Exception: %s", str(e))

            # Tidy our Named Temporary File
            _ntf_tidy(ntf)

            # Early Exit
            return False

        try:
            # write our content currently saved to disk to our temporary file
            with gzip.open(ntf.name, "wb") as f:
                # Write our content to disk
                f.write(
                    json.dumps(
                        {
                            k: v
                            for k, v in self._cache.items()
                            if v and v.persistent
                        },
                        separators=(",", ":"),
                        cls=CacheJSONEncoder,
                    ).encode(self.encoding)
                )

        except TypeError as e:
            # JSON object contains content that can not be encoded to disk
            logger.error(
                "Persistent temporary file can not be written to "
                "due to bad input data: %s",
                ntf.name,
            )
            logger.debug("Persistent Storage Exception: %s", str(e))

            # Tidy our Named Temporary File
            _ntf_tidy(ntf)

            # Early Exit
            return False

        except (OSError, EOFError, zlib.error) as e:
            logger.error(
                "Persistent temporary file inaccessible: %s", ntf.name
            )
            logger.debug("Persistent Storage Exception: %s", str(e))

            # Tidy our Named Temporary File
            _ntf_tidy(ntf)

            # Early Exit
            return False

        if not self.__move(ntf.name, cache_file):
            # Attempt to restore things as they were

            # Tidy our Named Temporary File
            _ntf_tidy(ntf)
            return False

        # Ensure our dirty flag is set to False
        self.__dirty = False

        return True

    def files(
        self,
        exclude: bool = True,
        lazy: bool = True,
    ) -> list[str]:
        """Returns the total files."""

        if lazy and exclude in self.__cache_files:
            # Take an early exit with our cached results
            return self.__cache_files[exclude]

        elif self.__mode == PersistentStoreMode.MEMORY:
            # Take an early exit
            # exclude is our cache switch and can be either True or False.
            # For the below, we just set both cases and set them up as an
            # empty record
            self.__cache_files.update({True: [], False: []})
            return []

        if not lazy or self.__exclude_list is None:
            # A list of criteria that should be excluded from the size count
            self.__exclude_list = (
                # Exclude backup cache file from count
                re.compile(
                    re.escape(
                        os.path.join(
                            self.__base_path,
                            f"{self.__cache_key}{self.__backup_extension}",
                        )
                    )
                ),
                # Exclude temporary files
                re.compile(re.escape(self.__temp_path) + r"[/\\].+"),
                # Exclude custom backup persistent files
                re.compile(
                    re.escape(self.__data_path)
                    + r"[/\\].+"
                    + re.escape(self.__backup_extension)
                ),
            )

        try:
            if exclude:
                self.__cache_files[exclude] = [
                    path
                    for path in filter(
                        os.path.isfile,
                        glob.glob(
                            os.path.join(self.__base_path, "**", "*"),
                            recursive=True,
                        ),
                    )
                    if next(
                        (False for p in self.__exclude_list if p.match(path)),
                        True,
                    )
                ]

            else:  # No exclusion list applied
                self.__cache_files[exclude] = list(
                    filter(
                        os.path.isfile,
                        glob.glob(
                            os.path.join(self.__base_path, "**", "*"),
                            recursive=True,
                        ),
                    )
                )

        except OSError:
            # We can't access the directory or it does not exist
            self.__cache_files[exclude] = []

        return self.__cache_files[exclude]

    @staticmethod
    def disk_scan(
        path: str,
        namespace: Optional[Union[str, list[str]]] = None,
        closest: bool = True,
    ) -> list[str]:
        """Scansk a path provided and returns namespaces detected."""

        logger.trace("Persistent path can of: %s", path)

        def is_namespace(x):
            """Validate what was detected is a valid namespace."""
            return os.path.isdir(
                os.path.join(path, x)
            ) and PersistentStore.__valid_key.match(x)

        # Handle our namespace searching
        if namespace:
            if isinstance(namespace, str):
                namespace = [namespace]

            elif not isinstance(namespace, (tuple, set, list)):
                raise AttributeError(
                    "namespace must be None, a string, or a tuple/set/list "
                    "of strings"
                )

        try:
            # Acquire all of the files in question
            namespaces = (
                [
                    ns
                    for ns in filter(is_namespace, os.listdir(path))
                    if not namespace
                    or next(
                        (True for n in namespace if ns.startswith(n)), False
                    )
                ]
                if closest
                else [
                    ns
                    for ns in filter(is_namespace, os.listdir(path))
                    if not namespace or ns in namespace
                ]
            )

        except FileNotFoundError:
            # no worries; Nothing to do
            logger.debug("Disk Prune path not found; nothing to clean.")
            return []

        except OSError as e:
            # Permission error of some kind or disk problem...
            # There is nothing we can do at this point
            logger.error("Disk Scan detetcted inaccessible path: %s", path)
            logger.debug("Persistent Storage Exception: %s", str(e))
            return []

        return namespaces

    @staticmethod
    def disk_prune(
        path: str,
        namespace: Optional[Union[str, list[str]]] = None,
        expires: Optional[Union[int, float]] = None,
        action: bool = False,
    ) -> dict[str, list[dict[str, Union[str, bool]]]]:
        """Prune persistent disk storage entries that are old and/or
        unreferenced.

        you must specify a path to perform the prune within

        if one or more namespaces are provided, then pruning focuses ONLY on
        those entries (if matched).

        if action is not set to False, directories to be removed are returned
        only
        """

        # Prepare our File Expiry
        expires = (
            datetime.now() - timedelta(seconds=expires)
            if isinstance(expires, (float, int)) and expires >= 0
            else PersistentStore.default_file_expiry
        )

        # Get our namespaces
        namespaces = PersistentStore.disk_scan(path, namespace)

        # Track matches
        _map = {}

        for namespace in namespaces:
            # Prepare our map
            _map[namespace] = []

            # Reference Directories
            base_dir = os.path.join(path, namespace)
            data_dir = os.path.join(base_dir, PersistentStore.data_dir)
            temp_dir = os.path.join(base_dir, PersistentStore.temp_dir)

            # Careful to only focus on files created by this Persistent Store
            # object
            files = [
                os.path.join(
                    base_dir,
                    f"{PersistentStore.__cache_key}"
                    f"{PersistentStore.__extension}",
                ),
                os.path.join(
                    base_dir,
                    f"{PersistentStore.__cache_key}"
                    f"{PersistentStore.__backup_extension}",
                ),
            ]

            # Update our files (applying what was defined above too)
            valid_data_re = re.compile(
                r".*("
                + re.escape(PersistentStore.__extension)
                + r"|"
                + re.escape(PersistentStore.__backup_extension)
                + r")$"
            )

            files = [
                path
                for path in filter(
                    os.path.isfile,
                    chain(
                        glob.glob(
                            os.path.join(data_dir, "*"), recursive=False
                        ),
                        files,
                    ),
                )
                if valid_data_re.match(path)
            ]

            # Now all temporary files
            files.extend(
                list(
                    filter(
                        os.path.isfile,
                        glob.glob(
                            os.path.join(temp_dir, "*"), recursive=False
                        ),
                    )
                )
            )

            # Track if we should do a directory sweep later on
            dir_sweep = True

            # Scan our files
            for file in files:
                try:
                    mtime = datetime.fromtimestamp(os.path.getmtime(file))

                except FileNotFoundError:
                    # no worries; we were removing it anyway
                    continue

                except OSError as e:
                    # Permission error of some kind or disk problem...
                    # There is nothing we can do at this point
                    logger.error(
                        "Disk Prune (ns=%s, clean=%s) detetcted inaccessible "
                        "file: %s",
                        namespace,
                        "yes" if action else "no",
                        file,
                    )
                    logger.debug("Persistent Storage Exception: %s", str(e))

                    # No longer worth doing a directory sweep
                    dir_sweep = False
                    continue

                if expires < mtime:
                    continue

                #
                # Handle Removing
                #
                record = {
                    "path": file,
                    "removed": False,
                }

                if action:
                    try:
                        os.unlink(file)
                        # Update our record
                        record["removed"] = True
                        logger.info(
                            "Disk Prune (ns=%s, clean=%s) removed persistent "
                            "file: %s",
                            namespace,
                            "yes" if action else "no",
                            file,
                        )

                    except FileNotFoundError:
                        # no longer worth doing a directory sweep
                        dir_sweep = False

                        # otherwise, no worries; we were removing the file
                        # anyway

                    except OSError as e:
                        # Permission error of some kind or disk problem...
                        # There is nothing we can do at this point
                        logger.error(
                            "Disk Prune (ns=%s, clean=%s) failed to remove "
                            "persistent file: %s",
                            namespace,
                            "yes" if action else "no",
                            file,
                        )

                        logger.debug(
                            "Persistent Storage Exception: %s", str(e)
                        )

                        # No longer worth doing a directory sweep
                        dir_sweep = False

                # Store our record
                _map[namespace].append(record)

            # Memory tidy
            del files

            if dir_sweep:
                # Gracefully cleanup our namespace directory. It's okay if we
                # fail; This just means there were files in the directory.
                for dirpath in (temp_dir, data_dir, base_dir):
                    if action:
                        try:
                            os.rmdir(dirpath)
                            logger.info(
                                "Disk Prune (ns=%s, clean=%s) removed "
                                "persistent dir: %s",
                                namespace,
                                "yes" if action else "no",
                                dirpath,
                            )
                        except OSError:
                            # do nothing;
                            pass
        return _map

    def size(
        self,
        exclude: bool = True,
        lazy: bool = True,
    ) -> int:
        """Returns the total size of the persistent storage in bytes."""

        if lazy and self.__cache_size is not None:
            # Take an early exit
            return self.__cache_size

        elif self.__mode == PersistentStoreMode.MEMORY:
            # Take an early exit
            self.__cache_size = 0
            return self.__cache_size

        # Get a list of files (file paths) in the given directory
        try:
            self.__cache_size = sum([
                os.stat(path).st_size
                for path in self.files(exclude=exclude, lazy=lazy)
            ])

        except OSError:
            # We can't access the directory or it does not exist
            self.__cache_size = 0

        return self.__cache_size

    def __del__(self) -> None:
        """Deconstruction of our object."""

        if self.__mode == PersistentStoreMode.AUTO:
            # Flush changes to disk
            self.flush()

    def __delitem__(self, key: str) -> None:
        """Remove a cache entry by it's key."""
        if self._cache is None and not self.__load_cache():
            raise KeyError("Could not initialize cache")

        try:
            if self._cache[key].persistent:
                # Set our dirty flag in advance
                self.__dirty = True

            # Store our new cache
            del self._cache[key]

        except KeyError:
            # Nothing to do
            raise

        if self.__dirty and self.__mode == PersistentStoreMode.FLUSH:
            # Flush changes to disk
            self.flush()

        return

    def __contains__(self, key: str) -> bool:
        """Verify if our storage contains the key specified or not.

        In additiont to this, if the content is expired, it is considered to be
        not contained in the storage.
        """
        if self._cache is None and not self.__load_cache():
            return False

        return key in self._cache and self._cache[key]

    def __setitem__(self, key: str, value: Any) -> None:
        """Sets a cache value without disrupting existing settings in place."""

        if self._cache is None and not self.__load_cache():
            raise KeyError("Could not initialize cache")

        if key not in self._cache and not self.set(key, value):
            raise KeyError("Could not set cache")

        else:
            # Update our value
            self._cache[key].set(value)

            if self._cache[key].persistent:
                # Set our dirty flag in advance
                self.__dirty = True

        if self.__dirty and self.__mode == PersistentStoreMode.FLUSH:
            # Flush changes to disk
            self.flush()

        return

    def __getitem__(self, key: str) -> Any:
        """Returns the indexed value."""

        if self._cache is None and not self.__load_cache():
            raise KeyError("Could not initialize cache")

        result = self.get(key, default=self.__not_found_ref, lazy=False)
        if result is self.__not_found_ref:
            raise KeyError(f" {key} not found in cache")

        return result

    def keys(self) -> builtins.set[str]:
        """Returns our keys."""
        if self._cache is None and not self.__load_cache():
            # There are no keys to return
            return {}.keys()

        return self._cache.keys()

    def delete(
        self,
        *args: str,
        all: Optional[bool] = None,
        temp: Optional[bool] = None,
        cache: Optional[bool] = None,
        validate: bool = True,
    ) -> bool:
        """Manages our file space and tidys it up.

        delete('key', 'key2') delete(all=True) delete(temp=True, cache=True)
        """

        # Our failure flag
        has_error = False

        valid_key_re = re.compile(
            r"^(?P<key>.+)("
            + re.escape(self.__backup_extension)
            + r"|"
            + re.escape(self.__extension)
            + r")$",
            re.I,
        )

        # Default asignments
        if all is None:
            all = bool(not (len(args) or temp or cache))
        if temp is None:
            temp = bool(all)
        if cache is None:
            cache = bool(all)

        if cache and self._cache:
            # Reset our object
            self._cache.clear()
            # Reset dirt flag
            self.__dirty = False

        for path in self.files(exclude=False):

            # Some information we use to validate the actions of our clean()
            # call. This is so we don't remove anything we shouldn't
            base = os.path.dirname(path)
            fname = os.path.basename(path)

            # Clean printable path details
            ppath = os.path.join(os.path.dirname(base), fname)

            if base == self.__base_path and cache:
                # We're handling a cache file (hopefully)
                result = valid_key_re.match(fname)
                key = (
                    None
                    if not result
                    else (
                        result["key"]
                        if self.__valid_key.match(result["key"])
                        else None
                    )
                )

                if validate and key != self.__cache_key:
                    # We're not dealing with a cache key
                    logger.debug(
                        "Persistent File cleanup ignoring file: %s", path
                    )
                    continue

                #
                # We should proceed with removing the file if we get here
                #

            elif base == self.__data_path and (args or all):
                # We're handling a file found in our custom data path
                result = valid_key_re.match(fname)
                key = (
                    None
                    if not result
                    else (
                        result["key"]
                        if self.__valid_key.match(result["key"])
                        else None
                    )
                )

                if validate and key is None:
                    # we're set to validate and a non-valid file was found
                    logger.debug(
                        "Persistent File cleanup ignoring file: %s", path
                    )
                    continue

                elif not all and (key is None or key not in args):
                    # no match found
                    logger.debug(
                        "Persistent File cleanup ignoring file: %s", path
                    )
                    continue

                #
                # We should proceed with removing the file if we get here
                #

            elif base == self.__temp_path and temp:
                #
                # This directory is a temporary path and nothing in here needs
                # to be further verified. Proceed with the removing of the file
                #
                pass

            else:
                # No match; move on
                logger.debug("Persistent File cleanup ignoring file: %s", path)
                continue

            try:
                os.unlink(path)
                logger.info("Removed persistent file: %s", ppath)

            except FileNotFoundError:
                # no worries; we were removing it anyway
                pass

            except OSError as e:
                # Permission error of some kind or disk problem...
                # There is nothing we can do at this point
                has_error = True
                logger.error("Failed to remove persistent file: %s", ppath)
                logger.debug("Persistent Storage Exception: %s", str(e))

        # Reset our reference variables
        self.__cache_size = None
        self.__cache_files.clear()

        return not has_error

    @property
    def cache_file(self) -> str:
        """Returns the full path to the namespace directory."""
        return os.path.join(
            self.__base_path,
            f"{self.__cache_key}{self.__extension}",
        )

    @property
    def path(self) -> Optional[str]:
        """Returns the full path to the namespace directory."""
        return self.__base_path

    @property
    def mode(self) -> PersistentStoreMode:
        """Returns the Persistent Storage mode."""
        return self.__mode



# BSD 2-Clause License
#
# Apprise - Push Notification Library.
# Copyright (c) 2025, Chris Caron <lead2gold@gmail.com>
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice,
#    this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

import copy
import os

from ..common import (
    NOTIFY_IMAGE_SIZES,
    NOTIFY_TYPES,
    NotifyImageSize,
    NotifyType,
)
from ..locale import LazyTranslation, gettext_lazy as _
from ..logger import logger
from ..manager_plugins import NotificationManager
from ..utils.cwe312 import cwe312_url
from ..utils.parse import GET_SCHEMA_RE, parse_list

# Used for testing
from .base import NotifyBase

# Grant access to our Notification Manager Singleton
N_MGR = NotificationManager()

__all__ = [
    "NOTIFY_IMAGE_SIZES",
    "NOTIFY_TYPES",
    "NotifyBase",
    # Reference
    "NotifyImageSize",
    "NotifyType",
    # Tokenizer
    "url_to_dict",
]


def _sanitize_token(tokens, default_delimiter):
    """This is called by the details() function and santizes the output by
    populating expected and consistent arguments if they weren't otherwise
    specified."""

    # Used for tracking groups
    group_map = {}

    # Iterate over our tokens
    for key in tokens:

        for element in tokens[key]:
            # Perform translations (if detected to do so)
            if isinstance(tokens[key][element], LazyTranslation):
                tokens[key][element] = str(tokens[key][element])

        if "alias_of" in tokens[key]:
            # Do not touch this field
            continue

        elif "name" not in tokens[key]:
            # Default to key
            tokens[key]["name"] = key

        if "map_to" not in tokens[key]:
            # Default type to key
            tokens[key]["map_to"] = key

        # Track our map_to objects
        if tokens[key]["map_to"] not in group_map:
            group_map[tokens[key]["map_to"]] = set()
        group_map[tokens[key]["map_to"]].add(key)

        if "type" not in tokens[key]:
            # Default type to string
            tokens[key]["type"] = "string"

        elif tokens[key]["type"].startswith("list"):
            if "delim" not in tokens[key]:
                # Default list delimiter (if not otherwise specified)
                tokens[key]["delim"] = default_delimiter

            if key in group_map[tokens[key]["map_to"]]:  # pragma: no branch
                # Remove ourselves from the list
                group_map[tokens[key]["map_to"]].remove(key)

            # Pointing to the set directly so we can dynamically update
            # ourselves
            tokens[key]["group"] = group_map[tokens[key]["map_to"]]

        elif (
            tokens[key]["type"].startswith("choice")
            and "default" not in tokens[key]
            and "values" in tokens[key]
            and len(tokens[key]["values"]) == 1
        ):

            # If there is only one choice; then make it the default
            #  - support dictionaries too
            tokens[key]["default"] = (
                tokens[key]["values"][0]
                if not isinstance(tokens[key]["values"], dict)
                else next(iter(tokens[key]["values"]))
            )

        if "values" in tokens[key] and isinstance(tokens[key]["values"], dict):
            # Convert values into a list if it was defined as a dictionary
            tokens[key]["values"] = list(tokens[key]["values"].keys())

        if "regex" in tokens[key]:
            # Verify that we are a tuple; convert strings to tuples
            if isinstance(tokens[key]["regex"], str):
                # Default tuple setup
                tokens[key]["regex"] = (tokens[key]["regex"], None)

            elif not isinstance(tokens[key]["regex"], (list, tuple)):
                # Invalid regex
                del tokens[key]["regex"]

        if "required" not in tokens[key]:
            # Default required is False
            tokens[key]["required"] = False

        if "private" not in tokens[key]:
            # Private flag defaults to False if not set
            tokens[key]["private"] = False
    return


def details(plugin):
    """Provides templates that can be used by developers to build URLs
    dynamically.

    If a list of templates is provided, then they will be used over the default
    value.

    If a list of tokens are provided, then they will over-ride any additional
    settings built from this script and/or will be appended to them afterwards.
    """

    # Our unique list of parsing will be based on the provided templates
    # if none are provided we will use our own
    templates = tuple(plugin.templates)

    # The syntax is simple
    #   {
    #       # The token_name must tie back to an entry found in the
    #       # templates list.
    #       'token_name': {
    #
    #            # types can be 'string', 'int', 'choice', 'list, 'float'
    #            # both choice and list may additionally have a : identify
    #            # what the list/choice type is comprised of; the default
    #            # is string.
    #            'type': 'choice:string',
    #
    #            # values will only exist the type must be a fixed
    #            # list of inputs (generated from type choice for example)
    #
    #            # If this is a choice:bool then you should ALWAYS define
    #            # this list as a (True, False) such as ('Yes, 'No') or
    #            # ('Enabled', 'Disabled'), etc
    #            'values': [ 'http', 'https' ],
    #
    #            # Identifies if the entry specified is required or not
    #            'required': True,
    #
    #            # Identifies all tokens detected to be associated with the
    #            # list:string
    #            # This is ony present in list:string objects and is only set
    #            # if this element acts as an alias for several other
    #            # kwargs/fields.
    #            'group': [],
    #
    #            # Identify a default value
    #            'default': 'http',
    #
    #            # Optional Verification Entries min and max are for floats
    #            # and/or integers
    #            'min': 4,
    #            'max': 5,
    #
    #            # A list will always identify a delimiter.  If this is
    #            # part of a path, this may be a '/', or it could be a
    #            # comma and/or space. delimiters are always in a list
    #            #  eg (if space and/or comma is a delimiter the entry
    #            #      would look like: 'delim': [',' , ' ' ]
    #            'delim': None,
    #
    #            # Use regex if you want to share the regular expression
    #            # required to validate the field. The regex will never
    #            # accomodate the prefix (if one is specified).  That is
    #            # up to the user building the URLs to include the prefix
    #            # on the URL when constructing it.
    #            # The format is ('regex', 'reg options')
    #            'regex': (r'[A-Z0-9]+', 'i'),
    #
    #            # A Prefix is always a string, to differentiate between
    #            # multiple arguments, sometimes content is prefixed.
    #            'prefix': '@',
    #
    #            # By default the key of this object is to be interpreted
    #            # as the argument to the notification in question. However
    #            # To accomodate cases where there are multiple types that
    #            # all map to the same entry, one can find a map_to value.
    #            'map_to': 'function_arg',
    #
    #            # Some arguments act as an alias_of an already defined object
    #            # This plays a role more with configuration file generation
    #            # since yaml files allow you to define different argumuments
    #            # in line to simplify things.  If this directive is set, then
    #            # it should be treated exactly the same as the object it is
    #            # an alias of
    #            'alias_of': 'function_arg',
    #
    #            # Advise developers to consider the potential sensitivity
    #            # of this field owned by the user. This is for passwords,
    #            # and api keys, etc...
    #            'private': False,
    #       },
    #   }

    # Template tokens identify the arguments required to initialize the
    # plugin itself.  It identifies all of the tokens and provides some
    # details on their use.  Each token defined should in some way map
    # back to at least one URL {token} defined in the templates

    # Since we nest a dictionary within a dictionary, a simple copy isn't
    # enough. a deepcopy allows us to manipulate this object in this
    # funtion without obstructing the original.
    template_tokens = copy.deepcopy(plugin.template_tokens)

    # Arguments and/or Options either have a default value and/or are
    # optional to be set.
    #
    # Since we nest a dictionary within a dictionary, a simple copy isn't
    # enough. a deepcopy allows us to manipulate this object in this
    # funtion without obstructing the original.
    template_args = copy.deepcopy(plugin.template_args)

    # Our template keyword arguments ?+key=value&-key=value
    # Basically the user provides both the key and the value. this is only
    # possibly by identifying the key prefix required for them to be
    # interpreted hence the +/- keys are built into apprise by default for easy
    # reference. In these cases, entry might look like '+' being the prefix:
    #   {
    #      'arg_name': {
    #          'name': 'label',
    #          'prefix': '+',
    #       }
    #   }
    #
    # Since we nest a dictionary within a dictionary, a simple copy isn't
    # enough. a deepcopy allows us to manipulate this object in this
    # funtion without obstructing the original.
    template_kwargs = copy.deepcopy(plugin.template_kwargs)

    # We automatically create a schema entry
    template_tokens["schema"] = {
        "name": _("Schema"),
        "type": "choice:string",
        "required": True,
        "values": parse_list(plugin.secure_protocol, plugin.protocol),
    }

    # Sanitize our tokens
    _sanitize_token(template_tokens, default_delimiter=("/",))
    # Delimiter(s) are space and/or comma
    _sanitize_token(template_args, default_delimiter=(",", " "))
    _sanitize_token(template_kwargs, default_delimiter=(",", " "))

    # Argument/Option Handling
    for key in list(template_args.keys()):

        if "alias_of" in template_args[key]:
            # Check if the mapped reference is a list; if it is, then
            # we need to store a different delimiter
            alias_of = template_tokens.get(template_args[key]["alias_of"], {})
            if (
                alias_of.get("type", "").startswith("list")
                and "delim" not in template_args[key]
            ):
                # Set a default delimiter of a comma and/or space if one
                # hasn't already been specified
                template_args[key]["delim"] = (",", " ")

        # _lookup_default looks up what the default value
        if "_lookup_default" in template_args[key]:
            template_args[key]["default"] = getattr(
                plugin, template_args[key]["_lookup_default"]
            )

            # Tidy as we don't want to pass this along in response
            del template_args[key]["_lookup_default"]

        # _exists_if causes the argument to only exist IF after checking
        # the return of an internal variable requiring a check
        if "_exists_if" in template_args[key]:
            if not getattr(plugin, template_args[key]["_exists_if"]):
                # Remove entire object
                del template_args[key]

            else:
                # We only nee to remove this key
                del template_args[key]["_exists_if"]

    return {
        "templates": templates,
        "tokens": template_tokens,
        "args": template_args,
        "kwargs": template_kwargs,
    }


def requirements(plugin):
    """Provides a list of packages and its requirement details."""
    requirements = {
        # Use the description to provide a human interpretable description of
        # what is required to make the plugin work. This is only nessisary
        # if there are package dependencies
        "details": "",
        # Define any required packages needed for the plugin to run.  This is
        # an array of strings that simply look like lines in the
        # `requirements.txt` file...
        #
        # A single string is perfectly acceptable:
        # 'packages_required' = 'cryptography'
        #
        # Multiple entries should look like the following
        # 'packages_required' = [
        #   'cryptography < 3.4`,
        # ]
        #
        "packages_required": [],
        # Recommended packages identify packages that are not required to make
        # your plugin work, but would improve it's use or grant it access to
        # full functionality (that might otherwise be limited).
        # Similar to `packages_required`, you would identify each entry in
        # the array as you would in a `requirements.txt` file.
        #
        #   - Do not re-provide entries already in the `packages_required`
        "packages_recommended": [],
    }

    # Populate our template differently if we don't find anything above
    if not (
        hasattr(plugin, "requirements")
        and isinstance(plugin.requirements, dict)
    ):
        # We're done early
        return requirements

    # Get our required packages
    _req_packages = plugin.requirements.get("packages_required")
    if isinstance(_req_packages, str):
        # Convert to list
        _req_packages = [_req_packages]

    elif not isinstance(_req_packages, (set, list, tuple)):
        # Allow one to set the required packages to None (as an example)
        _req_packages = []

    requirements["packages_required"] = [str(p) for p in _req_packages]

    # Get our recommended packages
    _opt_packages = plugin.requirements.get("packages_recommended")
    if isinstance(_opt_packages, str):
        # Convert to list
        _opt_packages = [_opt_packages]

    elif not isinstance(_opt_packages, (set, list, tuple)):
        # Allow one to set the recommended packages to None (as an example)
        _opt_packages = []

    requirements["packages_recommended"] = [str(p) for p in _opt_packages]

    # Get our package details
    _req_details = plugin.requirements.get("details")
    if not _req_details:
        if not (_req_packages or _opt_packages):
            _req_details = _("No dependencies.")

        elif _req_packages:
            _req_details = _("Packages are required to function.")

        else:  # opt_packages
            _req_details = _(
                "Packages are recommended to improve functionality."
            )
    else:
        # Store our details if defined
        requirements["details"] = _req_details

    # Return our compiled package requirements
    return requirements


def url_to_dict(url, secure_logging=True):
    """Takes an apprise URL and returns the tokens associated with it if they
    can be acquired based on the plugins available.

    None is returned if the URL could not be parsed, otherwise the tokens are
    returned.

    These tokens can be loaded into apprise through it's add() function.
    """

    # swap hash (#) tag values with their html version
    _url = url.replace("/#", "/%23")

    # CWE-312 (Secure Logging) Handling
    loggable_url = url if not secure_logging else cwe312_url(url)

    # Attempt to acquire the schema at the very least to allow our plugins to
    # determine if they can make a better interpretation of a URL geared for
    # them.
    schema = GET_SCHEMA_RE.match(_url)
    if schema is None:
        # Not a valid URL; take an early exit
        logger.error(f"Unsupported URL: {loggable_url}")
        return None

    # Ensure our schema is always in lower case
    schema = schema.group("schema").lower()
    if schema not in N_MGR:
        # Give the user the benefit of the doubt that the user may be using
        # one of the URLs provided to them by their notification service.
        # Before we fail for good, just scan all the plugins that support the
        # native_url() parse function
        results = None
        for plugin in N_MGR.plugins():
            results = plugin.parse_native_url(_url)
            if results:
                break

        if not results:
            logger.error(f"Unparseable URL {loggable_url}")
            return None

        logger.trace(
            "URL {} unpacked as:{}{}".format(
                url,
                os.linesep,
                os.linesep.join([f'{k}="{v}"' for k, v in results.items()]),
            )
        )

    else:
        # Parse our url details of the server object as dictionary
        # containing all of the information parsed from our URL
        results = N_MGR[schema].parse_url(_url)
        if not results:
            logger.error(
                f"Unparseable {N_MGR[schema].service_name} URL {loggable_url}"
            )
            return None

        logger.trace(
            "{} URL {} unpacked as:{}{}".format(
                N_MGR[schema].service_name,
                url,
                os.linesep,
                os.linesep.join([f'{k}="{v}"' for k, v in results.items()]),
            )
        )

    # Return our results
    return results



# BSD 2-Clause License
#
# Apprise - Push Notification Library.
# Copyright (c) 2025, Chris Caron <lead2gold@gmail.com>
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice,
#    this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

import requests

from ..common import NotifyFormat, NotifyType
from ..locale import gettext_lazy as _
from ..utils.parse import parse_bool, validate_regex
from .base import NotifyBase


class NotifyPushMe(NotifyBase):
    """A wrapper for PushMe Notifications."""

    # The default descriptive name associated with the Notification
    service_name = "PushMe"

    # The services URL
    service_url = "https://push.i-i.me/"

    # Insecure protocol (for those self hosted requests)
    protocol = "pushme"

    # A URL that takes you to the setup/help of the specific protocol
    setup_url = "https://github.com/caronc/apprise/wiki/Notify_pushme"

    # PushMe URL
    notify_url = "https://push.i-i.me/"

    # Define object templates
    templates = ("{schema}://{token}",)

    # Define our template tokens
    template_tokens = dict(
        NotifyBase.template_tokens,
        **{
            "token": {
                "name": _("Token"),
                "type": "string",
                "private": True,
                "required": True,
            },
        },
    )

    # Define our template arguments
    template_args = dict(
        NotifyBase.template_args,
        **{
            "token": {
                "alias_of": "token",
            },
            "push_key": {
                "alias_of": "token",
            },
            "status": {
                "name": _("Show Status"),
                "type": "bool",
                "default": True,
            },
        },
    )

    def __init__(self, token, status=None, **kwargs):
        """Initialize PushMe Object."""
        super().__init__(**kwargs)

        # Token (associated with project)
        self.token = validate_regex(token)
        if not self.token:
            msg = f"An invalid PushMe Token ({token}) was specified."
            self.logger.warning(msg)
            raise TypeError(msg)

        # Set Status type
        self.status = status

        return

    def send(self, body, title="", notify_type=NotifyType.INFO, **kwargs):
        """Perform PushMe Notification."""

        headers = {
            "User-Agent": self.app_id,
        }

        # Prepare our payload
        params = {
            "push_key": self.token,
            "title": (
                title
                if not self.status
                else f"{self.asset.ascii(notify_type)} {title}"
            ),
            "content": body,
            "type": (
                "markdown"
                if self.notify_format == NotifyFormat.MARKDOWN
                else "text"
            ),
        }

        self.logger.debug(
            "PushMe POST URL:"
            f" {self.notify_url} (cert_verify={self.verify_certificate!r})"
        )
        self.logger.debug(f"PushMe Payload: {params!s}")

        # Always call throttle before any remote server i/o is made
        self.throttle()

        try:
            r = requests.post(
                self.notify_url,
                params=params,
                headers=headers,
                verify=self.verify_certificate,
                timeout=self.request_timeout,
            )
            if r.status_code != requests.codes.ok:
                # We had a problem
                status_str = NotifyPushMe.http_response_code_lookup(
                    r.status_code
                )

                self.logger.warning(
                    "Failed to send PushMe notification:{}{}error={}.".format(
                        status_str, ", " if status_str else "", r.status_code
                    )
                )

                self.logger.debug(f"Response Details:\r\n{r.content}")

                # Return; we're done
                return False

            else:
                self.logger.info("Sent PushMe notification.")

        except requests.RequestException as e:
            self.logger.warning(
                "A Connection error occurred sending PushMe notification.",
            )
            self.logger.debug(f"Socket Exception: {e!s}")

            # Return; we're done
            return False

        return True

    @property
    def url_identifier(self):
        """Returns all of the identifiers that make this URL unique from
        another simliar one.

        Targets or end points should never be identified here.
        """
        return (self.secure_protocol, self.token)

    def url(self, privacy=False, *args, **kwargs):
        """Returns the URL built dynamically based on specified arguments."""

        # Define any URL parameters
        params = {
            "status": "yes" if self.status else "no",
        }

        # Extend our parameters
        params.update(self.url_parameters(privacy=privacy, *args, **kwargs))

        # Official URLs are easy to assemble
        return "{schema}://{token}/?{params}".format(
            schema=self.protocol,
            token=self.pprint(self.token, privacy, safe=""),
            params=NotifyPushMe.urlencode(params),
        )

    @staticmethod
    def parse_url(url):
        """Parses the URL and returns enough arguments that can allow us to re-
        instantiate this object."""
        results = NotifyBase.parse_url(url, verify_host=False)
        if not results:
            # We're done early as we couldn't load the results
            return results

        # Store our token using the host
        results["token"] = NotifyPushMe.unquote(results["host"])

        # The 'token' makes it easier to use yaml configuration
        if "token" in results["qsd"] and len(results["qsd"]["token"]):
            results["token"] = NotifyPushMe.unquote(results["qsd"]["token"])

        elif "push_key" in results["qsd"] and len(results["qsd"]["push_key"]):
            # Support 'push_key' if specified
            results["token"] = NotifyPushMe.unquote(results["qsd"]["push_key"])

        # Get status switch
        results["status"] = parse_bool(results["qsd"].get("status", True))

        return results



# BSD 2-Clause License
#
# Apprise - Push Notification Library.
# Copyright (c) 2025, Chris Caron <lead2gold@gmail.com>
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice,
#    this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

#
# API: https://github.com/Finb/bark-server/blob/master/docs/API_V2.md#python
#
import json

import requests

from ..common import NotifyImageSize, NotifyType
from ..locale import gettext_lazy as _
from ..url import PrivacyMode
from ..utils.parse import parse_bool, parse_list
from .base import NotifyBase

# Sounds generated off of: https://github.com/Finb/Bark/tree/master/Sounds
BARK_SOUNDS = (
    "alarm.caf",
    "anticipate.caf",
    "bell.caf",
    "birdsong.caf",
    "bloom.caf",
    "calypso.caf",
    "chime.caf",
    "choo.caf",
    "descent.caf",
    "electronic.caf",
    "fanfare.caf",
    "glass.caf",
    "gotosleep.caf",
    "healthnotification.caf",
    "horn.caf",
    "ladder.caf",
    "mailsent.caf",
    "minuet.caf",
    "multiwayinvitation.caf",
    "newmail.caf",
    "newsflash.caf",
    "noir.caf",
    "paymentsuccess.caf",
    "shake.caf",
    "sherwoodforest.caf",
    "silence.caf",
    "spell.caf",
    "suspense.caf",
    "telegraph.caf",
    "tiptoes.caf",
    "typewriters.caf",
    "update.caf",
)


# Supported Level Entries
class NotifyBarkLevel:
    """Defines the Bark Level options."""

    ACTIVE = "active"

    TIME_SENSITIVE = "timeSensitive"

    PASSIVE = "passive"

    CRITICAL = "critical"


BARK_LEVELS = (
    NotifyBarkLevel.ACTIVE,
    NotifyBarkLevel.TIME_SENSITIVE,
    NotifyBarkLevel.PASSIVE,
    NotifyBarkLevel.CRITICAL,
)


class NotifyBark(NotifyBase):
    """A wrapper for Notify Bark Notifications."""

    # The default descriptive name associated with the Notification
    service_name = "Bark"

    # The services URL
    service_url = "https://github.com/Finb/Bark"

    # The default protocol
    protocol = "bark"

    # The default secure protocol
    secure_protocol = "barks"

    # A URL that takes you to the setup/help of the specific protocol
    setup_url = "https://github.com/caronc/apprise/wiki/Notify_bark"

    # Allows the user to specify the NotifyImageSize object; this is supported
    # through the webhook
    image_size = NotifyImageSize.XY_128

    # Define object templates
    templates = (
        "{schema}://{host}/{targets}",
        "{schema}://{host}:{port}/{targets}",
        "{schema}://{user}:{password}@{host}/{targets}",
        "{schema}://{user}:{password}@{host}:{port}/{targets}",
    )

    # Define our template arguments
    template_tokens = dict(
        NotifyBase.template_tokens,
        **{
            "host": {
                "name": _("Hostname"),
                "type": "string",
                "required": True,
            },
            "port": {
                "name": _("Port"),
                "type": "int",
                "min": 1,
                "max": 65535,
            },
            "user": {
                "name": _("Username"),
                "type": "string",
            },
            "password": {
                "name": _("Password"),
                "type": "string",
                "private": True,
            },
            "target_device": {
                "name": _("Target Device"),
                "type": "string",
                "map_to": "targets",
            },
            "targets": {
                "name": _("Targets"),
                "type": "list:string",
                "required": True,
            },
        },
    )

    # Define our template arguments
    template_args = dict(
        NotifyBase.template_args,
        **{
            "to": {
                "alias_of": "targets",
            },
            "sound": {
                "name": _("Sound"),
                "type": "choice:string",
                "values": BARK_SOUNDS,
            },
            "level": {
                "name": _("Level"),
                "type": "choice:string",
                "values": BARK_LEVELS,
            },
            "volume": {
                "name": _("Volume"),
                "type": "int",
                "min": 0,
                "max": 10,
            },
            "click": {
                "name": _("Click"),
                "type": "string",
            },
            "badge": {
                "name": _("Badge"),
                "type": "int",
                "min": 0,
            },
            "category": {
                "name": _("Category"),
                "type": "string",
            },
            "group": {
                "name": _("Group"),
                "type": "string",
            },
            "image": {
                "name": _("Include Image"),
                "type": "bool",
                "default": True,
                "map_to": "include_image",
            },
            "icon": {
                "name": _("Icon URL"),
                "type": "string",
            },
            "call": {
                "name": _("Call"),
                "type": "bool",
                "default": False,
            },
        },
    )

    def __init__(
        self,
        targets=None,
        include_image=True,
        sound=None,
        category=None,
        group=None,
        level=None,
        click=None,
        badge=None,
        volume=None,
        icon=None,
        call=None,
        **kwargs,
    ):
        """Initialize Notify Bark Object."""
        super().__init__(**kwargs)

        # Prepare our URL
        self.notify_url = "{}://{}{}/push".format(
            "https" if self.secure else "http",
            self.host,
            (
                f":{self.port}"
                if (self.port and isinstance(self.port, int))
                else ""
            ),
        )

        # Assign our category
        self.category = category if isinstance(category, str) else None

        # Assign our group
        self.group = group if isinstance(group, str) else None

        # Initialize device list
        self.targets = parse_list(targets)

        # Place an image inline with the message body
        self.include_image = include_image

        # A clickthrough option for notifications
        self.click = click

        # Badge
        try:
            # Acquire our badge count if we can:
            #  - We accept both the integer form as well as a string
            #    representation
            self.badge = int(badge)
            if self.badge < 0:
                raise ValueError()

        except TypeError:
            # NoneType means use Default; this is an okay exception
            self.badge = None

        except ValueError:
            self.badge = None
            self.logger.warning(
                "The specified Bark badge ({}) is not valid ", badge
            )

        # Sound (easy-lookup)
        self.sound = (
            None
            if not sound
            else next(
                (f for f in BARK_SOUNDS if f.startswith(sound.lower())), None
            )
        )
        if sound and not self.sound:
            self.logger.warning(
                "The specified Bark sound ({}) was not found ", sound
            )

        # Volume
        self.volume = None
        if volume is not None:
            try:
                self.volume = int(volume) if volume is not None else None
                if self.volume is not None and not (0 <= self.volume <= 10):
                    raise ValueError()

            except (TypeError, ValueError):
                self.logger.warning(
                    "The specified Bark volume ({}) is not valid. "
                    "Must be between 0 and 10",
                    volume,
                )

        # Call
        self.call = parse_bool(call)

        # Icon URL
        self.icon = icon if isinstance(icon, str) else None

        # Level
        self.level = (
            None
            if not level
            else next((f for f in BARK_LEVELS if f[0] == level[0]), None)
        )
        if level and not self.level:
            self.logger.warning(
                "The specified Bark level ({}) is not valid ", level
            )

        return

    def send(self, body, title="", notify_type=NotifyType.INFO, **kwargs):
        """Perform Bark Notification."""

        # error tracking (used for function return)
        has_error = False

        if not self.targets:
            # We have nothing to notify; we're done
            self.logger.warning("There are no Bark devices to notify")
            return False

        # Prepare our headers
        headers = {
            "User-Agent": self.app_id,
            "Content-Type": "application/json; charset=utf-8",
        }

        # Prepare our payload (sample below)
        # {
        #     "body": "Test Bark Server",
        #     "device_key": "nysrshcqielvoxsa",
        #     "title": "bleem",
        #     "category": "category",
        #     "sound": "minuet.caf",
        #     "badge": 1,
        #     "icon": "https://day.app/assets/images/avatar.jpg",
        #     "group": "test",
        #     "url": "https://mritd.com"
        # }
        payload = {
            "title": title if title else self.app_desc,
            "body": body,
        }

        # Acquire our image url if configured to do so
        image_url = (
            None if not self.include_image else self.image_url(notify_type)
        )

        # Use custom icon if provided, otherwise use default image
        if self.icon:
            payload["icon"] = self.icon
        elif image_url:
            payload["icon"] = image_url

        if self.sound:
            payload["sound"] = self.sound

        if self.click:
            payload["url"] = self.click

        if self.badge:
            payload["badge"] = self.badge

        if self.level:
            payload["level"] = self.level

        if self.category:
            payload["category"] = self.category

        if self.group:
            payload["group"] = self.group

        if self.volume:
            payload["volume"] = self.volume

        if self.call:
            payload["call"] = 1

        auth = None
        if self.user:
            auth = (self.user, self.password)

        # Create a copy of the targets
        targets = list(self.targets)

        while len(targets) > 0:
            # Retrieve our device key
            target = targets.pop()

            payload["device_key"] = target
            self.logger.debug(
                "Bark POST URL:"
                f" {self.notify_url} (cert_verify={self.verify_certificate!r})"
            )
            self.logger.debug(f"Bark Payload: {payload!s}")

            # Always call throttle before any remote server i/o is made
            self.throttle()
            try:
                r = requests.post(
                    self.notify_url,
                    data=json.dumps(payload),
                    headers=headers,
                    auth=auth,
                    verify=self.verify_certificate,
                    timeout=self.request_timeout,
                )
                if r.status_code != requests.codes.ok:
                    # We had a problem
                    status_str = NotifyBark.http_response_code_lookup(
                        r.status_code
                    )

                    self.logger.warning(
                        "Failed to send Bark notification to {}: "
                        "{}{}error={}.".format(
                            target,
                            status_str,
                            ", " if status_str else "",
                            r.status_code,
                        )
                    )

                    self.logger.debug(f"Response Details:\r\n{r.content}")

                    # Mark our failure
                    has_error = True
                    continue

                else:
                    self.logger.info(f"Sent Bark notification to {target}.")

            except requests.RequestException as e:
                self.logger.warning(
                    "A Connection error occurred sending Bark "
                    f"notification to {target}."
                )
                self.logger.debug(f"Socket Exception: {e!s}")

                # Mark our failure
                has_error = True
                continue

        return not has_error

    @property
    def url_identifier(self):
        """Returns all of the identifiers that make this URL unique from
        another simliar one.

        Targets or end points should never be identified here.
        """
        return (
            self.secure_protocol if self.secure else self.protocol,
            self.user,
            self.password,
            self.host,
            self.port,
        )

    def url(self, privacy=False, *args, **kwargs):
        """Returns the URL built dynamically based on specified arguments."""

        # Define any URL parameters
        params = {
            "image": "yes" if self.include_image else "no",
        }

        if self.sound:
            params["sound"] = self.sound

        if self.click:
            params["click"] = self.click

        if self.badge:
            params["badge"] = str(self.badge)

        if self.level:
            params["level"] = self.level

        if self.volume:
            params["volume"] = str(self.volume)

        if self.category:
            params["category"] = self.category

        if self.group:
            params["group"] = self.group

        if self.icon:
            params["icon"] = self.icon

        if self.call:
            params["call"] = "yes"

        # Extend our parameters
        params.update(self.url_parameters(privacy=privacy, *args, **kwargs))

        # Determine Authentication
        auth = ""
        if self.user and self.password:
            auth = "{user}:{password}@".format(
                user=NotifyBark.quote(self.user, safe=""),
                password=self.pprint(
                    self.password, privacy, mode=PrivacyMode.Secret, safe=""
                ),
            )
        elif self.user:
            auth = "{user}@".format(
                user=NotifyBark.quote(self.user, safe=""),
            )

        default_port = 443 if self.secure else 80
        return "{schema}://{auth}{hostname}{port}/{targets}?{params}".format(
            schema=self.secure_protocol if self.secure else self.protocol,
            auth=auth,
            # never encode hostname since we're expecting it to be a valid one
            hostname=self.host,
            port=(
                ""
                if self.port is None or self.port == default_port
                else f":{self.port}"
            ),
            targets="/".join([NotifyBark.quote(f"{x}") for x in self.targets]),
            params=NotifyBark.urlencode(params),
        )

    def __len__(self):
        """Returns the number of targets associated with this notification."""
        return len(self.targets)

    @staticmethod
    def parse_url(url):
        """Parses the URL and returns enough arguments that can allow us to re-
        instantiate this object."""

        results = NotifyBase.parse_url(url)
        if not results:
            # We're done early as we couldn't load the results
            return results

        # Apply our targets
        results["targets"] = NotifyBark.split_path(results["fullpath"])

        # Category
        if "category" in results["qsd"] and results["qsd"]["category"]:
            results["category"] = NotifyBark.unquote(
                results["qsd"]["category"].strip()
            )

        # Group
        if "group" in results["qsd"] and results["qsd"]["group"]:
            results["group"] = NotifyBark.unquote(
                results["qsd"]["group"].strip()
            )

        # Badge
        if "badge" in results["qsd"] and results["qsd"]["badge"]:
            results["badge"] = NotifyBark.unquote(
                results["qsd"]["badge"].strip()
            )

        # Volume
        if "volume" in results["qsd"] and results["qsd"]["volume"]:
            results["volume"] = NotifyBark.unquote(
                results["qsd"]["volume"].strip()
            )

        # Level
        if "level" in results["qsd"] and results["qsd"]["level"]:
            results["level"] = NotifyBark.unquote(
                results["qsd"]["level"].strip()
            )

        # Click (URL)
        if "click" in results["qsd"] and results["qsd"]["click"]:
            results["click"] = NotifyBark.unquote(
                results["qsd"]["click"].strip()
            )

        # Sound
        if "sound" in results["qsd"] and results["qsd"]["sound"]:
            results["sound"] = NotifyBark.unquote(
                results["qsd"]["sound"].strip()
            )

        # The 'to' makes it easier to use yaml configuration
        if "to" in results["qsd"] and len(results["qsd"]["to"]):
            results["targets"] += NotifyBark.parse_list(results["qsd"]["to"])

        # use image= for consistency with the other plugins
        results["include_image"] = parse_bool(
            results["qsd"].get("image", True)
        )

        # Icon URL
        if "icon" in results["qsd"] and results["qsd"]["icon"]:
            results["icon"] = NotifyBark.unquote(
                results["qsd"]["icon"].strip()
            )

        # Call
        results["call"] = parse_bool(
            results["qsd"].get("call", False)
        )

        return results
